-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity rv_exec is
  generic ( RISCV_M    : sl_t := '0';        -- RV mul / div extension
            RISCV_Zba  : sl_t := '0' );      -- address generation (shXadd)
  port ( clk_i           : in  std_logic;
         xrst_i          : in  std_logic;
         --
         opcode_p2_i     : in  slv_31_00;
         pc_e1_i         : in  slv_31_00;
         hold_i          : in  std_logic;
         --
         opcode_valid_i  : in  std_logic;
         opcode_p1_i     : in  slv_31_00;    -- (branch imm12, bimm, jimm20)
         opcode_pc_i     : in  slv_31_00;
         oprs1_dat_i     : in  slv_31_00;
         oprs2_dat_i     : in  slv_31_00;
         --  alu
         alu_op_i        : in  alu_op_t;
         wb_value_o      : out slv_31_00;
         --  multiplier
         mul_op_valid_i  : in  std_logic;
         mul_op_p1_i     : in  mul_op_t;
         mul_hold_i      : in  std_logic;
         mul_wb_value_o  : out slv_31_00;
         --  divider
         div_op_valid_i  : in  std_logic;
         div_op_p1_i     : in  div_op_t;
         div_wb_valid_o  : out std_logic;
         div_wb_value_o  : out slv_31_00;
         --  branch
         branch_op_i     : in  branch_op_t;
         branch_req_o    : out std_logic;
         branch_pc_o     : out slv_31_00
         );
end entity rv_exec;


architecture rtl of rv_exec is

  component rv_alu
    port ( alu_op_i     : in  slv_03_00;
           alu_a_i      : in  slv_31_00;
           alu_b_i      : in  slv_31_00;
           alu_p_o      : out slv_31_00 );
  end component;


  component rv_multiplier
    port ( clk_i        : in  std_logic;
           xrst_i       : in  std_logic;
           --
           mul_op_p1_i  : in  mul_op_t;
           valid_i      : in  std_logic;
           rs1_dat_i    : in  slv_31_00;
           rs2_dat_i    : in  slv_31_00;
           hold_i       : in  std_logic;
           --
           wb_value_o   : out slv_31_00 );
  end component;


  component rv_divider
    port ( clk_i        : in  std_logic;
           xrst_i       : in  std_logic;
           --
           div_op_p1_i  : in  div_op_t;
           valid_i      : in  std_logic;
           rs1_dat_i    : in  slv_31_00;
           rs2_dat_i    : in  slv_31_00;
           --
           wb_valid_o   : out std_logic;
           wb_value_o   : out slv_31_00 );
  end component;


  --  opcode decode
  signal imm12_w      : slv_31_00;
  signal bimm_w       : slv_31_00;
  signal jimm20_w     : slv_31_00;
  signal func3_w      : slv_t(FUNC3_R'length-1 downto 0);

  --  execute - ALU operations
  signal alu_a_w      : slv_31_00;
  signal alu_b_w      : slv_31_00;

  --  ALU
  signal alu_dout_w   : slv_31_00;

  --  execute - branch operations
  signal branch_w     : std_logic;
  signal branch_go_w  : std_logic;
  signal branch_pc_w  : slv_31_00;

  signal oprs1_dat_q  : slv_31_00;
  signal oprs2_dat_q  : slv_31_00;
  signal instr_p2     : alu_op_t;

  signal imm20_q      : slv_31_00;
  signal imm12_q      : slv_31_00;

  signal mul_dout_w   : slv_31_00;
  signal div_dout_w   : slv_31_00;

begin


  -- MULHSU | MULH          :  a  <= ra_i(31)  &  ra_i(31:0)
  -- MULHU  | MUL           :  a  <= '0'       &  ra_i(31:0)
  --
  -- MULH                   :  b  <= rb_i(31)  &  rb_i(31:0)
  -- MULHU  | MUL | MULHSU  :  b  <= '0'       &  rb_i(31:0)
  u_rv_multiplier : rv_multiplier
    port map ( clk_i        => clk_i,
               xrst_i       => xrst_i,
               -- inputs
               mul_op_p1_i  => mul_op_p1_i,       -- <<dec
               valid_i      => mul_op_valid_i,    -- <<iss
               rs1_dat_i    => oprs1_dat_i,       -- <<iss
               rs2_dat_i    => oprs2_dat_i,       -- <<iss
               hold_i       => mul_hold_i,        -- <<iss
               -- outputs
               wb_value_o   => mul_dout_w );      -- >>iss

  mul_wb_value_o <= mul_dout_w  when ( RISCV_M = '1' ) else X"00000000";



  -- ra : dividend[31:0]
  -- rb : divisor[62:0]
  --
  -- DIV | REM ==> signed op
  --
  -- (DIV | REM) & ra_i(31) = '1'   :  dividend[31:0]  <=  -ra_i
  -- else (DIVU | REMU | !ra_i(31)) :  dividend[31:0]  <=   ra_i
  --
  -- (DIV | REM) & rb_i(31) = '1'   :  divisor[62:0]   <= -rb_i &  (30:0 => '0')
  -- else (DIVU | REMU | !rb_i(31)) :  divisor[62:0]   <=  rb_i &  (30:0 => '0')
  u_rv_divider : rv_divider
    port map ( clk_i        => clk_i,
               xrst_i       => xrst_i,
               -- inputs
               div_op_p1_i  => div_op_p1_i,       -- <<dec
               valid_i      => div_op_valid_i,    -- <<iss
               rs1_dat_i    => oprs1_dat_i,       -- <<iss
               rs2_dat_i    => oprs2_dat_i,       -- <<iss
               -- outputs
               wb_valid_o   => div_wb_valid_o,    -- >>iss
               wb_value_o   => div_dout_w );      -- >>iss

  div_wb_value_o <= div_dout_w  when ( RISCV_M = '1' ) else X"00000000";
  --
  -- invert divider output when
  --   invert_res <= '1' :   DIV & (ra__i(31) /= rb_i(31)) and rb_i(31:0) /= 0
  --                       or  REM and ra_i(31) = '1'
  -- i.e. inver output when
  --  1) DIV op  and  ra/rb are different sign and rb is not zero vector
  --  2) REM op  and  ra    is negative (dividend)




  -- - - - - - - - - - - - -
  --  opcode decode
  -- - - - - - - - - - - - -
  instr_p2   <= alu_op_i;

  imm20_q     <= ( opcode_p2_i(IMM20_R'range) & X"000" );        -- 31:12
  imm12_q     <= ( (19 downto 0 => opcode_p2_i(IMM12_R'left)) &  -- sign
                   opcode_p2_i(IMM12_R'range) );                 -- 31:20


  p_aluireg : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        oprs1_dat_q <= (others => '0');
        oprs2_dat_q <= (others => '0');
      else
        if (hold_i = '0') then
          oprs1_dat_q <= oprs1_dat_i;
          oprs2_dat_q <= oprs2_dat_i;
        end if;
      end if;
    end if;
  end process p_aluireg;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- execute - ALU operations
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --
  -- select ALU A input
  --
  --   NOTE:  load upper immediate is in b-bus (instr_p2.i_lui, less mux)
  --   oprs1_dat_q : alu op / add immediate
  --   pc_e1_i     : add upper immediate to pc / jump and link [register]
  --
  --   Zba - address generation: sh1add / sh2add / sh3add
  --
  alu_a_w <= oprs1_dat_q(30 downto 0) & '0'   when (instr_p2.i_sh1add = '1')
             else
             oprs1_dat_q(29 downto 0) & "00"  when (instr_p2.i_sh2add = '1')
             else
             oprs1_dat_q(28 downto 0) & "000" when (instr_p2.i_sh3add = '1')
             else
             oprs1_dat_q when (instr_p2.i_2operand = '1' or
                               instr_p2.i_immediate = '1')  else
             pc_e1_i     when ( instr_p2.i_auipc = '1' or
                                instr_p2.i_jal = '1'   or
                                instr_p2.i_jalr = '1' ) else
             X"00000000";


  --
  -- select ALU B input
  --
  --   oprs2_dat_q : alu op / add immediate / sh*add
  --   imm12_q     : immediate op, (shamt_q : shift amount (slli, srli, srai))
  --   imm20_q     : ! ! load upper immediate, NOTE: ON B BUS
  --   X"00000004" : jal / jalr: (indirect jump, jump and link register)
  alu_b_w <= oprs2_dat_q when (instr_p2.i_2operand = '1') else
             imm12_q when (instr_p2.i_immediate = '1')    else  -- immediate add
             imm20_q when (instr_p2.i_lui = '1' or instr_p2.i_auipc = '1')  else
             X"00000004" when (instr_p2.i_jal = '1' or instr_p2.i_jalr = '1')
             else X"00000000";



  u_alu : rv_alu
    port map ( alu_op_i => instr_p2.alu_func4,
               alu_a_i  => alu_a_w,     -- oprs1_dat_q / pc_e1_i / X"00000000"
               alu_b_i  => alu_b_w,     -- rs2_dat_q / imm-- / shamt / +4
               alu_p_o  => alu_dout_w );


  wb_value_o   <= alu_dout_w;



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  branch
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  imm12_w     <= ( (19 downto 0 => opcode_p1_i(31)) &   -- 4 branch
                   opcode_p1_i(31 downto 20) );
  bimm_w      <= ( (18 downto 0 => opcode_p1_i(31)) &   -- 4 branch
                   opcode_p1_i(31) &
                   opcode_p1_i(7) &
                   opcode_p1_i(30 downto 25) &
                   opcode_p1_i(11 downto 8) & '0' );
  jimm20_w    <= ( (11 downto 0 => opcode_p1_i(31)) &   -- 4 branch
                   opcode_p1_i(19 downto 12) &
                   opcode_p1_i(20) &
                   opcode_p1_i(30 downto 25) &
                   opcode_p1_i(24 downto 21) & '0' );
  
  --
  -- default branch_w target is relative to current PC
  --      : addu(opcode_pc_i, bimm_w;
  -- jal  : addu(opcode_pc_i, jimm20_w)
  -- jalr : adds(oprs1_dat_i, imm12_w)
  --
  p_c_bpc : process( branch_op_i, opcode_pc_i, bimm_w, jimm20_w, oprs1_dat_i,
                     imm12_w)
    variable arg1_v : slv_t(31+1 downto 0);
    variable arg2_v : slv_t(31+1 downto 0);
  begin
    --  default branch_w target is relative to current PC
    arg1_v := '0' & opcode_pc_i;                      -- UNSIGNED ADD
    arg2_v := '0' & bimm_w;
    if ( branch_op_i.i_jal = '1' ) then               -- jal
      arg2_v := '0' & jimm20_w;
    end if;
    if ( branch_op_i.i_jalr = '1' ) then              -- jump and link register
      arg1_v := oprs1_dat_i(oprs1_dat_i'left)  &  oprs1_dat_i;    -- SIGNED ADD
      arg2_v := imm12_w(imm12_w'left)          &  imm12_w;
    end if;
    branch_pc_w     <= adds(arg1_v, arg2_v)(31 downto 1) & '0';
  end process p_c_bpc;


  branch_go_w <= branch_op_i.i_jal or
                 branch_op_i.i_jalr or
                 ( branch_op_i.i_beq and eq(oprs1_dat_i, oprs2_dat_i) ) or
                 (branch_op_i.i_bne and (not (eq(oprs1_dat_i, oprs2_dat_i)))) or
                 (branch_op_i.i_blt and lt_signed(oprs1_dat_i,  oprs2_dat_i)) or
                 (branch_op_i.i_bge and
                  (not lt_signed(oprs1_dat_i, oprs2_dat_i))) or   -- = GE
                 (branch_op_i.i_bltu and     lt(oprs1_dat_i, oprs2_dat_i)) or
                 (branch_op_i.i_bgeu and not lt(oprs1_dat_i, oprs2_dat_i));


  branch_w    <= branch_op_i.i_jal or branch_op_i.i_jalr or
                 branch_op_i.i_beq or
                 branch_op_i.i_bne or branch_op_i.i_blt or
                 branch_op_i.i_bge or
                 branch_op_i.i_bltu or branch_op_i.i_bgeu;

  -- branch signals exec -> issue
  branch_req_o <= opcode_valid_i and branch_w and branch_go_w;
  branch_pc_o  <= branch_pc_w;                                    -- 31:00


end rtl;
