-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity rv_issue is
  generic ( RISCV_M         : std_logic := '1';  -- RV mul / div extension
            RISCV_E         : std_logic := '0';  -- RV rf embedded extension
            USE_LD_BYPASS   : std_logic := '1';
            USE_MUL_BYPASS  : std_logic := '1' );
  port ( clk_i                      : in  std_logic;
         xrst_i                     : in  std_logic;  -- syncr reset
         --
         fetch_valid_i              : in  std_logic;
         --  instr from decode
         fetch_instr_i              : in  slv_31_00;
         fetch_pc_i                 : in  slv_31_00;
         fetch_fault_fetch_i        : in  std_logic;
         fetch_fault_page_i         : in  std_logic;
         -- instruction group (csr, mul, div, ...)
         idec_group_i               : in  i_group_t;
         iss_stall_o                : out std_logic;
         iss_step_o                 : out std_logic;
         --
         branch_exe_req_i           : in  std_logic;
         branch_exe_pc_i            : in  slv_31_00;
         branch_csr_req_i           : in  std_logic;
         branch_csr_pc_i            : in  slv_31_00;
         branch_csr_priv_i          : in  slv_01_00;
         --
         mem_ld_fault_i             : in  std_logic;
         mem_st_fault_i             : in  std_logic;
         mem_ack_i                  : in  std_logic;
         mem_error_i                : in  std_logic;
         dmem_din_i                 : in  slv_31_00;
         dmem_ldtype_i              : in dmem_ldtyp_t;
         --
         exec_opcode_valid_o        : out std_logic;
         exec_hold_o                : out std_logic;
         wb_exec_value_i            : in  slv_31_00;  -- ALU
         -- signals to / from divider
         div_opcode_valid_o         : out std_logic;
         wb_div_valid_i             : in  std_logic;
         wb_div_value_i             : in  slv_31_00;  -- DIV
         -- signals to / from multiplier
         mul_opcode_valid_o         : out std_logic;  -- MUL
         mul_hold_o                 : out std_logic;
         wb_mul_value_i             : in  slv_31_00;
         --
         csr_result_e1_value_i      : in  slv_31_00;
         csr_result_e1_write_i      : in  std_logic;
         csr_result_e1_wdata_i      : in  slv_31_00;
         csr_result_e1_exception_i  : in  slv_05_00;
         lsu_stall_i                : in  std_logic;
         take_interrupt_i           : in  std_logic;
         --  outputs
         fetch_accept_o             : out std_logic;
         branch_request_o           : out std_logic;
         branch_pc_o                : out slv_31_00;
         branch_priv_o              : out slv_01_00;
         lsu_opcode_valid_o         : out std_logic;
         csr_opcode_valid_o         : out std_logic;
         --
         opcode_p2_o                : out slv_31_00;
         pc_e1_o                    : out slv_31_00;
         oprs1_idx_o                : out slv_04_00;
         oprs1_dat_o                : out slv_31_00;
         oprs2_dat_o                : out slv_31_00;
         csr_opcode_invalid_o       : out std_logic;
         csr_wb_write_o             : out std_logic;
         csr_wb_waddr_o             : out slv_11_00;
         csr_wb_wdata_o             : out slv_31_00;
         csr_wb_exception_o         : out slv_05_00;
         csr_wb_exception_pc_o      : out slv_31_00;
         csr_wb_exception_addr_o    : out slv_31_00;
         interrupt_inhibit_o        : out std_logic
         );
end entity rv_issue;


architecture rtl of rv_issue is

  component rv_pipe_ctrl
    generic ( USE_LD_BYPASS   : std_logic := '1';
              USE_MUL_BYPASS  : std_logic := '1' );
    port ( clk_i                      : in  std_logic;
           xrst_i                     : in  std_logic;   -- syncr reset
           --  issue
           issue_valid_i              : in  std_logic;
           issue_accept_i             : in  std_logic;
           issue_stall_i              : in  std_logic;
           issue_lsu_i                : in  std_logic;
           issue_csr_i                : in  std_logic;
           issue_csr_reg_i            : in  std_logic;   -- reg rd/wr
           issue_div_i                : in  std_logic;
           issue_mul_i                : in  std_logic;
           issue_branch_i             : in  std_logic;
           issue_rd_valid_i           : in  std_logic;
           issue_rd_i                 : in  slv_04_00;
           issue_exception_i          : in  slv_05_00;
           take_interrupt_i           : in  std_logic;
           issue_branch_taken_i       : in  std_logic;
           issue_branch_target_i      : in  slv_31_00;
           issue_pc_i                 : in  slv_31_00;
           issue_opcode_i             : in  slv_31_00;
           issue_operand_ra_i         : in  slv_31_00;
           issue_operand_rb_i         : in  slv_31_00;
           --  execution stage 1: ALU result
           alu_result_e1_i            : in  slv_31_00;
           --  execution stage 1: CSR read result / early exceptions
           csr_result_value_e1_i      : in  slv_31_00;
           csr_result_write_e1_i      : in  std_logic;
           csr_result_wdata_e1_i      : in  slv_31_00;
           csr_result_exception_e1_i  : in  slv_05_00;
           --  execution stage 1
           load_e1_o                  : out std_logic;
           store_e1_o                 : out std_logic;
           mul_e1_o                   : out std_logic;
           rd_e1_o                    : out slv_04_00;
           pc_e1_o                    : out slv_31_00;
           --  execution stage 2 : other results
           mem_ld_fault_i             : in  std_logic;
           mem_st_fault_i             : in  std_logic;
           mem_ack_i                  : in  std_logic;
           mem_err_i                  : in  std_logic;
           dmem_din_i                 : in  slv_31_00;
           dmem_ldtype_i              : in  dmem_ldtyp_t;
           mul_result_e2_i            : in  slv_31_00;
           --  execution stage 2
           load_e2_o                  : out std_logic;
           mul_e2_o                   : out std_logic;
           rd_e2_o                    : out slv_04_00;
           result_e2_o                : out slv_31_00;
           --
           opcode_p2_o                : out slv_31_00;
           --  out of pipe  : divide result
           div_complete_i             : in  std_logic;
           div_result_i               : in  slv_31_00;
           --  commit
           csr_wb_o                   : out std_logic;
           rd_wb_o                    : out slv_04_00;
           result_wb_o                : out slv_31_00;
           pc_wb_o                    : out slv_31_00;
           operand_ra_wb_o            : out slv_31_00;
           exception_wb_o             : out slv_05_00;
           csr_write_wb_o             : out std_logic;
           csr_waddr_wb_o             : out slv_11_00;
           csr_wdata_wb_o             : out slv_31_00;
           --
           stall_o                    : out std_logic;
           squash_e1_e2_o             : out std_logic );
  end component;

  component rv_regfile
    generic ( RISCV_E  : std_logic := '0' );       -- RV rf embedded extension
    port ( clk_i      : in  std_logic;
           xrst_i     : in  std_logic;
           --  write port
           rd_addr_i  : in  slv_04_00;
           rd_data_i  : in  slv_31_00;
           -- read port a
           ra_addr_i  : in  slv_04_00;
           ra_data_o  : out slv_31_00;
           --  read port b
           rb_addr_i  : in  slv_04_00;
           rb_data_o  : out slv_31_00 );
  end component;


  signal stall_w               : std_logic;
  signal squash_w              : std_logic;

  --  issue select
  signal opcode_valid_w        : std_logic;

  signal rs1_idx_w             : slv_04_00;
  signal rs2_idx_w             : slv_04_00;
  signal rd_idx_w              : slv_04_00;
  signal issue_sb_alloc_w      : std_logic;
  signal issue_lsu_w           : std_logic;
  signal issue_branch_w        : std_logic;
  signal issue_mul_w           : std_logic;
  signal issue_div_w           : std_logic;
  signal issue_csr_w           : std_logic;
  signal issue_invalid_w       : std_logic;

  --  pipeline status tracking
  signal pipe_squash_e1_e2_w   : std_logic;

  signal opcode_accept_w       : std_logic;
  signal pipe_stall_raw_w      : std_logic;

  signal pipe_load_e1_w        : std_logic;
  signal pipe_store_e1_w       : std_logic;
  signal pipe_mul_e1_w         : std_logic;
  signal pipe_rd_e1_w          : slv_04_00;

  signal pipe_load_e2_w        : std_logic;
  signal pipe_mul_e2_w         : std_logic;
  signal pipe_rd_e2_w          : slv_04_00;
  signal pipe_result_e2_w      : slv_31_00;

  signal pipe_csr_wb_w         : std_logic;
  signal pipe_rd_wb_w          : slv_04_00;
  signal pipe_result_wb_w      : slv_31_00;
  signal pipe_pc_wb_w          : slv_31_00;
  signal pipe_exception_wb_w   : slv_t(EXCEPTION_W-1 downto 0);

  signal issue_fault_w         : slv_t(EXCEPTION_W-1 downto 0);

  --  blocking events (division, CSR unit access)
  signal div_pending_q         : std_logic;
  signal csr_pending_q         : std_logic;

  --  register file
  signal rf_rs1_dat_w          : slv_31_00;
  signal rf_rs2_dat_w          : slv_31_00;

  signal issue_a_dat_w         : slv_31_00;
  signal issue_b_dat_w         : slv_31_00;

  --  cannot read output
  signal oprs1_dat_o_w         : slv_31_00;
  signal oprs2_dat_o_w         : slv_31_00;
  signal div_opcode_valid_o_w  : std_logic;
  signal csr_opcode_valid_o_w  : std_logic;

  signal ra_e2_sel_w           : std_logic;
  signal rb_e2_sel_w           : std_logic;

begin

  -- - - - - - - - - - - - -
  --  instruction Decoder
  -- - - - - - - - - - - - -
  rs1_idx_w        <= fetch_instr_i(RS1_R'range);    -- 19:15 rf src 1
  rs2_idx_w        <= fetch_instr_i(RS2_R'range);    -- 24:20 rf src 2
  rd_idx_w         <= fetch_instr_i(RD_R'range);     -- 11:7  rf destination

  issue_sb_alloc_w <= idec_group_i.rd_valid;

  issue_lsu_w      <= idec_group_i.lsu;
  issue_branch_w   <= idec_group_i.branch;
  issue_mul_w      <= idec_group_i.mul;
  issue_div_w      <= idec_group_i.div;
  issue_csr_w      <= idec_group_i.csr;
  issue_invalid_w  <= idec_group_i.invalid;


  issue_fault_w <= EXCEPTION_FAULT_FETCH when (fetch_fault_fetch_i = '1') else
                   EXCEPTION_PAGE_FAULT_INST when (fetch_fault_page_i = '1')
                   else (EXCEPTION_W-1 downto 0 => '0');


  -- - - - - - - - - - - - -
  --  issue select
  -- - - - - - - - - - - - -
  opcode_valid_w   <= fetch_valid_i and (not squash_w) and
                      (not branch_csr_req_i);

  --  branch request (CSR branch - ecall, xret, or branch instruction)
  branch_request_o <= branch_csr_req_i or branch_exe_req_i;
  branch_pc_o      <= branch_csr_pc_i when (branch_csr_req_i = '1')
                      else branch_exe_pc_i;
  branch_priv_o    <= branch_csr_priv_i;

  iss_stall_o      <= stall_w;
  iss_step_o       <= opcode_accept_w and (not pipe_squash_e1_e2_w);


  u_pipe_ctrl : rv_pipe_ctrl
    generic map ( USE_LD_BYPASS   => USE_LD_BYPASS,
                  USE_MUL_BYPASS  => USE_MUL_BYPASS )
    port map ( clk_i                      => clk_i,
               xrst_i                     => xrst_i,
               --  issue
               issue_valid_i              => opcode_accept_w,
               issue_accept_i             => opcode_accept_w,
               issue_stall_i              => stall_w,
               issue_lsu_i                => issue_lsu_w,
               issue_csr_i                => issue_csr_w,
               issue_csr_reg_i            => idec_group_i.csr_reg, -- reg rd/wr
               issue_div_i                => issue_div_w,
               issue_mul_i                => issue_mul_w,
               issue_branch_i             => issue_branch_w,
               issue_rd_valid_i           => issue_sb_alloc_w,
               issue_rd_i                 => rd_idx_w,
               issue_exception_i          => issue_fault_w,
               issue_pc_i                 => fetch_pc_i,
               issue_opcode_i             => fetch_instr_i,
               issue_operand_ra_i         => oprs1_dat_o_w,
               issue_operand_rb_i         => oprs2_dat_o_w,
               issue_branch_taken_i       => branch_exe_req_i,
               issue_branch_target_i      => branch_exe_pc_i,
               take_interrupt_i           => take_interrupt_i,
               --  execution stage 1: ALU result
               alu_result_e1_i            => wb_exec_value_i,
               csr_result_value_e1_i      => csr_result_e1_value_i,
               csr_result_write_e1_i      => csr_result_e1_write_i,
               csr_result_wdata_e1_i      => csr_result_e1_wdata_i,
               csr_result_exception_e1_i  => csr_result_e1_exception_i,
               --  execution stage 1
               --
               opcode_p2_o                => opcode_p2_o,
               pc_e1_o                    => pc_e1_o,
               --
               load_e1_o                  => pipe_load_e1_w,
               store_e1_o                 => pipe_store_e1_w,
               mul_e1_o                   => pipe_mul_e1_w,
               rd_e1_o                    => pipe_rd_e1_w,
               --  execution stage 2: Other results
               mem_ld_fault_i             => mem_ld_fault_i,
               mem_st_fault_i             => mem_st_fault_i,
               mem_ack_i                  => mem_ack_i,
               mem_err_i                  => mem_error_i,
               dmem_din_i                 => dmem_din_i,
               dmem_ldtype_i              => dmem_ldtype_i,
               mul_result_e2_i            => wb_mul_value_i,
               --  execution stage 2
               load_e2_o                  => pipe_load_e2_w,
               mul_e2_o                   => pipe_mul_e2_w,
               rd_e2_o                    => pipe_rd_e2_w,
               result_e2_o                => pipe_result_e2_w, -- BB mem/mul
               stall_o                    => pipe_stall_raw_w,
               squash_e1_e2_o             => pipe_squash_e1_e2_w,
               --  out of pipe: divide result
               div_complete_i             => wb_div_valid_i,
               div_result_i               => wb_div_value_i,
               --  commit
               csr_wb_o                   => pipe_csr_wb_w,
               rd_wb_o                    => pipe_rd_wb_w,
               result_wb_o                => pipe_result_wb_w,
               pc_wb_o                    => pipe_pc_wb_w,
               operand_ra_wb_o            => OPEN,
               exception_wb_o             => pipe_exception_wb_w,
               csr_write_wb_o             => csr_wb_write_o,
               csr_waddr_wb_o             => csr_wb_waddr_o,
               csr_wdata_wb_o             => csr_wb_wdata_o );


  exec_hold_o  <= stall_w;
  mul_hold_o   <= stall_w;


  -- - - - - - - - - - - - -
  --  pipe1 - status tracking
  -- - - - - - - - - - - - -
  csr_wb_exception_o      <= pipe_exception_wb_w;
  csr_wb_exception_pc_o   <= pipe_pc_wb_w;
  csr_wb_exception_addr_o <= pipe_result_wb_w;


  -- - - - - - - - - - - - - - - - - - - - - - -
  --  blocking events (division, CSR unit access)
  -- - - - - - - - - - - - - - - - - - - - - - -

  --  division operations take 32+ (or 16+) cycles and stall
  --  the pipeline (complete out-of-pipe) until completed
  p_div_pend : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        div_pending_q <= '0';
      else
        if (pipe_squash_e1_e2_w = '1') then
          div_pending_q <= '0';
        elsif (div_opcode_valid_o_w = '1' and issue_div_w = '1') then
          div_pending_q <= '1';
        elsif (wb_div_valid_i = '1') then
          div_pending_q <= '0';
        end if;
      end if;
    end if;
  end process p_div_pend;


  --  CSR operations are infrequent - avoid any complications of pipelining
  --  them. These only take a 2-3 cycles anyway and may result in a pipe flush
  --  (e.g. ecall, ebreak..)
  p_csr_pend : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        csr_pending_q <= '0';
      else
        if (pipe_squash_e1_e2_w = '1') then
          csr_pending_q <= '0';
        elsif (csr_opcode_valid_o_w = '1' and issue_csr_w = '1') then
          csr_pending_q <= '1';
        elsif (pipe_csr_wb_w = '1') then
          csr_pending_q <= '0';
        end if;
      end if;
    end if;
  end process p_csr_pend;


  squash_w <= pipe_squash_e1_e2_w;


  -- - - - - - - - - - - - -
  --  issue / scheduling logic
  -- - - - - - - - - - - - -
  p_c_scoreb : process (pipe_load_e2_w, pipe_rd_e2_w, pipe_mul_e2_w,
                        pipe_load_e1_w, pipe_mul_e1_w, pipe_rd_e1_w,
                        pipe_store_e1_w, issue_mul_w, issue_div_w,
                        issue_csr_w, lsu_stall_i, stall_w,
                        div_pending_q, csr_pending_q, opcode_valid_w,
                        rs1_idx_w, rs2_idx_w, rd_idx_w)
    variable scoreb_v : slv_31_00;
  begin
    scoreb_v         := X"00000000";
    opcode_accept_w  <= '0';
    --
    --  execution units with >= 2 cycle latency
    scoreb_v(to_uint(pipe_rd_e2_w)) := pipe_load_e2_w and (not USE_LD_BYPASS);
    scoreb_v(to_uint(pipe_rd_e2_w)) := pipe_mul_e2_w and (not USE_MUL_BYPASS);
    --
    --  execution units with >= 1 cycle latency (loads / multiply)
    scoreb_v(to_uint(pipe_rd_e1_w)) := pipe_load_e1_w or pipe_mul_e1_w;
    --
    --  do not start multiply, division or CSR operation in the cycle
    --  after a load (leaving only ALU operations and branches)
    if ((pipe_load_e1_w = '1' or pipe_store_e1_w = '1') and
        (issue_mul_w = '1' or issue_div_w  = '1' or issue_csr_w = '1')) then
    scoreb_v := X"FFFFFFFF";
    end if;
    --
    --   stall - no issues...
    if (lsu_stall_i = '1' or stall_w = '1' or div_pending_q = '1' or
        csr_pending_q = '1') then
      null;
    --   primary slot (lsu, branch, alu, mul, div, csr)
    elsif (opcode_valid_w  = '1' and
           (scoreb_v(to_uint(rs1_idx_w)) = '0' and
            scoreb_v(to_uint(rs2_idx_w)) = '0' and
            scoreb_v(to_uint(rd_idx_w)) = '0') ) then
      opcode_accept_w <= '1';
    end if;
  end process p_c_scoreb;



  lsu_opcode_valid_o    <= opcode_accept_w and (not take_interrupt_i);
  exec_opcode_valid_o   <= opcode_accept_w;

  div_opcode_valid_o_w  <= opcode_accept_w  and RISCV_M;
  mul_opcode_valid_o    <= div_opcode_valid_o_w;
  div_opcode_valid_o    <= div_opcode_valid_o_w;

  interrupt_inhibit_o   <= csr_pending_q or issue_csr_w;

  fetch_accept_o        <= (opcode_accept_w and
                            (not take_interrupt_i)) when (opcode_valid_w = '1')
                           else '1';

  stall_w               <= pipe_stall_raw_w;



  -- - - - - - - - - - - - -
  --  register file
  -- - - - - - - - - - - - -
  --  register file: 1W2R
  u_regfile : rv_regfile
    generic map ( RISCV_E => RISCV_E )      -- RV rf embedded extension
    port map (clk_i       => clk_i,
              xrst_i      => xrst_i,
              --  write port
              rd_addr_i   => pipe_rd_wb_w,
              rd_data_i   => pipe_result_wb_w,
              --  read port a
              ra_addr_i   => rs1_idx_w,
              ra_data_o   => rf_rs1_dat_w,
              --  read port b
              rb_addr_i   => rs2_idx_w,
              rb_data_o   => rf_rs2_dat_w );




  p_c_issue : process (rf_rs1_dat_w, rf_rs2_dat_w, pipe_rd_wb_w,
                       rs1_idx_w, pipe_result_wb_w, rs2_idx_w,
                       pipe_rd_e1_w, wb_exec_value_i)
  begin
    --  NOTE: newest version of operand takes priority
    issue_a_dat_w <= rf_rs1_dat_w;
    issue_b_dat_w <= rf_rs2_dat_w;

    --  bypass - WB
    if (rs1_idx_w = pipe_rd_wb_w) then
        issue_a_dat_w <= pipe_result_wb_w;
    end if;
    if (rs2_idx_w = pipe_rd_wb_w) then
        issue_b_dat_w <= pipe_result_wb_w;
    end if;
    --  bypass - E1
    if (rs1_idx_w = pipe_rd_e1_w) then
        issue_a_dat_w <= wb_exec_value_i;    -- mul/div/alu
    end if;
    if (rs2_idx_w = pipe_rd_e1_w) then
        issue_b_dat_w <= wb_exec_value_i;    -- mul/div/alu
    end if;
    --  reg 0 source
    if ( is_zero(rs1_idx_w) ) then           -- 5b
        issue_a_dat_w <= X"00000000";        -- 32b
    end if;
    if ( is_zero(rs2_idx_w) ) then           -- 5b
        issue_b_dat_w <= X"00000000";        -- 32b
    end if;
  end process p_c_issue;



  ra_e2_sel_w <= eq(rs1_idx_w, pipe_rd_e2_w) and
                 ne(rs1_idx_w, pipe_rd_e1_w) and
                 not_zero(rs1_idx_w);

  rb_e2_sel_w <= eq(rs2_idx_w, pipe_rd_e2_w) and
                 ne(rs2_idx_w, pipe_rd_e1_w) and
                 not_zero(rs2_idx_w);


  -- - - - - - - - - - - - -
  --  issue slot 0
  -- - - - - - - - - - - - -
  -- dmem / mul / pipe.reg : alu / csr
  oprs1_dat_o_w         <= pipe_result_e2_w when (ra_e2_sel_w = '1')
                           else issue_a_dat_w;
  oprs2_dat_o_w         <= pipe_result_e2_w when (rb_e2_sel_w = '1')
                           else issue_b_dat_w;

  oprs1_idx_o           <= rs1_idx_w;        -- reg addr rs1

  oprs1_dat_o           <= oprs1_dat_o_w;
  oprs2_dat_o           <= oprs2_dat_o_w;

  csr_opcode_valid_o_w  <= opcode_accept_w and (not take_interrupt_i);
  csr_opcode_valid_o    <= csr_opcode_valid_o_w;
  csr_opcode_invalid_o  <= opcode_accept_w and issue_invalid_w;


end rtl;
