-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity rv_regfile is
  generic ( RISCV_E : std_logic := '0' );   -- RV rf embedded extension
  port ( clk_i      : in  std_logic;
         xrst_i     : in  std_logic;        -- syncr reset
         --  write port
         rd_addr_i  : in  slv_04_00;
         rd_data_i  : in  slv_31_00;
         -- read port a
         ra_addr_i  : in  slv_04_00;
         ra_data_o  : out slv_31_00;
         -- read port b
         rb_addr_i  : in  slv_04_00;
         rb_data_o  : out slv_31_00
         );
end entity rv_regfile;


architecture rtl of rv_regfile is

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- flop based register file
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- Register   ABI Name    Description                       Saver
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- x28-31     t3-6        Temporaries                       Caller
  -- x18-27     s2-11       Saved registers                   Callee
  -- x12-17     a2-7        Function arguments                Caller
  -- x10-11     a0-1        Function Arguments/return values  Caller
  -- x9         s1          Saved register                    Callee
  -- x8         s0/fp       Saved register/Frame pointer      Callee
  -- x5-7       t0-2        Temporaries                       Caller
  -- x4         tp          Thread pointer                    -
  -- x3         gp          Global pointer                    -
  -- x2         sp          Stack pointer                     Callee
  -- x1         ra          Return address                    Caller
  -- x0         zero        Hard-wired zero                   -
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  -- pragma translate_off
  -- simulation friendly names
  signal x31_t6_w   : slv_31_00;
  signal x30_t5_w   : slv_31_00;
  signal x29_t4_w   : slv_31_00;
  signal x28_t3_w   : slv_31_00;
  signal x27_s11_w  : slv_31_00;
  signal x26_s10_w  : slv_31_00;
  signal x25_s9_w   : slv_31_00;
  signal x24_s8_w   : slv_31_00;
  signal x23_s7_w   : slv_31_00;
  signal x22_s6_w   : slv_31_00;
  signal x21_s5_w   : slv_31_00;
  signal x20_s4_w   : slv_31_00;
  signal x19_s3_w   : slv_31_00;
  signal x18_s2_w   : slv_31_00;
  signal x17_a7_w   : slv_31_00;
  signal x16_a6_w   : slv_31_00;
  signal x15_a5_w   : slv_31_00;
  signal x14_a4_w   : slv_31_00;
  signal x13_a3_w   : slv_31_00;
  signal x12_a2_w   : slv_31_00;
  signal x11_a1_w   : slv_31_00;
  signal x10_a0_w   : slv_31_00;
  signal x9_s1_w    : slv_31_00;
  signal x8_s0_w    : slv_31_00;
  signal x7_t2_w    : slv_31_00;
  signal x6_t1_w    : slv_31_00;
  signal x5_t0_w    : slv_31_00;
  signal x4_tp_w    : slv_31_00;
  signal x3_gp_w    : slv_31_00;
  signal x2_sp_w    : slv_31_00;
  signal x1_ra_w    : slv_31_00;
  signal x0_zero_w  : slv_31_00;
  -- pragma translate_on


  -- one hot write enebles
  signal wr_ohot_w  : slv_31_00;
  signal rfile_q    : rfile_t;

begin

  -- pragma translate_off
  -- simulation friendly names
  x31_t6_w   <= rfile_q(31);
  x30_t5_w   <= rfile_q(30);
  x29_t4_w   <= rfile_q(29);
  x28_t3_w   <= rfile_q(28);   -- t3-6    temporaries
  x27_s11_w  <= rfile_q(27);
  x26_s10_w  <= rfile_q(26);
  x25_s9_w   <= rfile_q(25);
  x24_s8_w   <= rfile_q(24);
  x23_s7_w   <= rfile_q(23);
  x22_s6_w   <= rfile_q(22);
  x21_s5_w   <= rfile_q(21);
  x20_s4_w   <= rfile_q(20);
  x19_s3_w   <= rfile_q(19);
  x18_s2_w   <= rfile_q(18);   -- s2-11   saved registers
  x17_a7_w   <= rfile_q(17);
  x16_a6_w   <= rfile_q(16);
  x15_a5_w   <= rfile_q(15);
  x14_a4_w   <= rfile_q(14);
  x13_a3_w   <= rfile_q(13);
  x12_a2_w   <= rfile_q(12);   -- a2-7    function arguments
  x11_a1_w   <= rfile_q(11);
  x10_a0_w   <= rfile_q(10);   -- a0-1    function arguments/return values
  x9_s1_w    <= rfile_q(9);    -- s1      saved register
  x8_s0_w    <= rfile_q(8);    -- s0/fp   saved register/Frame pointer
  x7_t2_w    <= rfile_q(7);
  x6_t1_w    <= rfile_q(6);
  x5_t0_w    <= rfile_q(5);    -- t0-2    temporaries
  x4_tp_w    <= rfile_q(4);    -- tp      thread pointer
  x3_gp_w    <= rfile_q(3);    -- gp      global pointer
  x2_sp_w    <= rfile_q(2);    -- sp      stack pointer
  x1_ra_w    <= rfile_q(1);    -- ra      return address
  x0_zero_w  <= rfile_q(0);    -- zero    hard-wired zero   (32b x '0')
  -- pragma translate_on



  -- - - - - - - - - - - - - - - - - -
  --  flop based register file
  -- - - - - - - - - - - - - - - - - -

  p_one_hot_rf32(rd_addr_i, RISCV_E, wr_ohot_w);

  p_dff32x32(clk_i, xrst_i, wr_ohot_w, rd_data_i, rfile_q);

  p_rf32x2mux(ra_addr_i, rb_addr_i, rfile_q, ra_data_o, rb_data_o);

end rtl;
