-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity dbg_bridge is
  generic ( CLK_FREQ    : natural := 50000000;         -- default 50M
            UART_SPEED  : natural :=  1000000 );       -- default 1M
  port ( clk_i          : in  std_logic;
         xrst_i         : in  std_logic;
         --  AXI bus to DDR / TCM ram if
         mem_awready_i  : in  std_logic;
         mem_wready_i   : in  std_logic;
         mem_bvalid_i   : in  std_logic;
         mem_arready_i  : in  std_logic;
         mem_rvalid_i   : in  std_logic;
         mem_rdata_i    : in  slv_31_00;
         mem_awvalid_o  : out std_logic;
         mem_awaddr_o   : out slv_31_00;
         mem_wvalid_o   : out std_logic;
         mem_wdata_o    : out slv_31_00;
         mem_wstrb_o    : out slv_03_00;
         mem_bready_o   : out std_logic;
         mem_arvalid_o  : out std_logic;
         mem_araddr_o   : out slv_31_00;
         mem_rready_o   : out std_logic;
         --  pp io
    --   gpio_i         : in  slv_31_00;               -- can be used for debug
    --   gpio_o         : out slv_31_00;               -- can be used for debug
         uart_rxd_i     : in  std_logic;
         uart_txd_o     : out std_logic;
         rv_core_ena_o  : out std_logic                -- cpu / cache enable
         );
end entity dbg_bridge;


architecture rtl of dbg_bridge is

  -- - - - - - - - - - - - -
  --  config
  -- - - - - - - - - - - - -
  -- bit_div_i is calculated as:
  --    slv_t(to_unsigned(CLK_FREQ/UART_SPEED, UART_DIVISOR_W))
  --
  -- e.g. IM uart at 50MHz system clock gives
  --  50e6 / 1M  =  50  =  0x32
  --
  
  constant UART_DIVISOR_W : natural := log2(CLK_FREQ/UART_SPEED) + 1;
  constant UART_DIVIDER   : slv_t(UART_DIVISOR_W-1 downto 0) :=
    slv_t(to_unsigned((CLK_FREQ/UART_SPEED)-1, UART_DIVISOR_W));
--  slv_t(to_unsigned(CLK_FREQ/UART_SPEED, UART_DIVISOR_W));


  constant GPIO_ADDRESS  : slv_31_00 := X"F0000000";
  constant STS_ADDRESS   : slv_31_00 := X"F0000004";

  -- - - - - - - - - - - - -
  --  constants
  -- - - - - - - - - - - - -
  constant REQ_WRITE     : slv_07_00  := X"10";
  constant REQ_READ      : slv_07_00  := X"11";

  constant STATE_W       : natural                       := 4;
  constant STATE_R       : slv_03_00  := X"0";  -- range def

  constant STATE_IDLE    : slv_03_00  := X"0";
  constant STATE_LEN     : slv_03_00  := X"2";
  constant STATE_ADDR0   : slv_03_00  := X"3";
  constant STATE_ADDR1   : slv_03_00  := X"4";
  constant STATE_ADDR2   : slv_03_00  := X"5";
  constant STATE_ADDR3   : slv_03_00  := X"6";
  constant STATE_WRITE   : slv_03_00  := X"7";
  constant STATE_READ    : slv_03_00  := X"8";
  constant STATE_DATA0   : slv_03_00  := X"9";
  constant STATE_DATA1   : slv_03_00  := X"A";
  constant STATE_DATA2   : slv_03_00  := X"B";
  constant STATE_DATA3   : slv_03_00  := X"C";


  component dbg_bridge_uart
    generic ( UART_DIVISOR_W  : natural := 9 );
    port ( clk_i       : in  std_logic;
           xrst_i      : in  std_logic;
           --  control
           bit_div_i   : in  slv_t(UART_DIVISOR_W-1 downto 0);
           stop_bits_i : in  std_logic;        -- 0 = 1, 1 = 2
           --  transmit
           wr_i        : in  std_logic;
           data_i      : in  slv_07_00;
           tx_busy_o   : out std_logic;
           --  receive
           rd_i        : in  std_logic;
           data_o      : out slv_07_00;
           rx_ready_o  : out std_logic;
           rx_err_o    : out std_logic;
           --  UART pins
           rxd_i       : in  std_logic;
           txd_o       : out std_logic );
  end component;


  component dbg_bridge_fifo
    generic ( WIDTH   : natural := 8;
              DEPTH   : natural := 4;
              ADDR_W  : natural := 2 );
    port ( clk_i      : in  std_logic;
           xrst_i     : in  std_logic;
           --  inputs
           data_in_i  : in  slv_t(WIDTH-1 downto 0);
           push_i     : in  std_logic;
           pop_i      : in  std_logic;
           --  outputs
           data_out_o : out slv_t(WIDTH-1 downto 0);
           accept_o   : out std_logic;
           valid_o    : out std_logic );
  end component;

  -- - - - - - - - - - - - -
  --  registers / wires
  -- - - - - - - - - - - - -

  --  states
  signal state_q          : slv_t(STATE_R'range);
  signal next_state_w     : slv_t(STATE_R'range);

  signal uart_wr_w        : std_logic;
  signal uart_wr_data_w   : slv_07_00;
  signal uart_wr_busy_w   : std_logic;

  signal uart_rd_w        : std_logic;
  signal uart_rd_data_w   : slv_07_00;
  signal uart_rd_valid_w  : std_logic;

  signal tx_valid_w       : std_logic;
  signal tx_data_w        : slv_07_00;
  signal tx_accept_w      : std_logic;
  signal read_skip_w      : std_logic;

  signal rx_valid_w       : std_logic;
  signal rx_data_w        : slv_07_00;
  signal rx_accept_w      : std_logic;

  signal mem_addr_q       : slv_31_00;
  signal mem_busy_q       : std_logic;
  signal mem_wr_q         : std_logic;

  signal len_q            : slv_07_00;

  --  byte index
  signal data_idx_q       : slv_01_00;

  --  word storage
  signal data_q           : slv_31_00;
  --  output FIFO
  -- - - - - - - - - - - - -
  signal uart_tx_pop_w    : std_logic;

  --  AXI: write request
  signal  mem_awvalid_q   : std_logic;
  signal  mem_awvalid_w   : std_logic;
                          
  signal  mem_wvalid_q    : std_logic;
  signal  mem_wvalid_w    : std_logic;

  --  AXI: read request
  signal mem_arvalid_q    : std_logic;
  signal mem_arvalid_w    : std_logic;

  --  write mask
  signal mem_sel_q        : slv_03_00;
  signal mem_sel_w        : slv_03_00;

  --  DBG
  signal magic_addr_w     : std_logic;

  --  cannot read output
  signal mem_awvalid_o_w  : std_logic;
  signal mem_wvalid_o_w   : std_logic;
  signal mem_arvalid_o_w  : std_logic;

  --  GPIO
  signal gpio_in_w        : slv_31_00;
  signal gpio_wr_q        : std_logic;
  signal gpio_out_q       : slv_31_00;
--signal gpio_out_w       : slv_31_00;

begin


  -- - - - - - - - - - - - -
  --  GPIO out
  -- - - - - - - - - - - - -
  p_gpwr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        gpio_wr_q    <= '0';
      else
        if (mem_addr_q = GPIO_ADDRESS and state_q = STATE_WRITE and
            rx_valid_w = '1' and
            (data_idx_q = "11" or unsigned(len_q) = 1)) then
          gpio_wr_q  <= '1';
        else
          gpio_wr_q  <= '0';
        end if;
      end if;
    end if;
  end process p_gpwr;

  p_gpo : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        gpio_out_q    <= X"00000000";
      else
        if (gpio_wr_q = '1') then
          gpio_out_q  <= data_q(31 downto 0);
        end if;
      end if;
    end if;
  end process p_gpo;


  rv_core_ena_o  <= gpio_out_q(0);                        -- core xreset
--gpio_out_w     <= gpio_out_q;
  -- gpio_o      <= gpio_out_q;                           -- for debug

  gpio_in_w      <= X"FACE" & gpio_out_q(15 downto 0);
  -- gpio_in_w   <= gpio_i;                               -- for debug

  
  --  DBG
  magic_addr_w <= '1' when ( mem_addr_q = GPIO_ADDRESS or
                             mem_addr_q = STS_ADDRESS ) else '0';



  -- - - - - - - - - - - - -
  --  UART core
  -- - - - - - - - - - - - -
  u_uart : dbg_bridge_uart
    generic map ( UART_DIVISOR_W => UART_DIVISOR_W )
    port map ( clk_i        => clk_i,
               xrst_i       => xrst_i,
               --  control
               bit_div_i    => UART_DIVIDER,
               stop_bits_i  => '0',                       -- 0 = 1, 1 = 2
               --  transmit
               wr_i         => uart_wr_w,
               data_i       => uart_wr_data_w,
               tx_busy_o    => uart_wr_busy_w,
               --  receive
               rd_i         => uart_rd_w,
               data_o       => uart_rd_data_w,
               rx_ready_o   => uart_rd_valid_w,
               rx_err_o     => OPEN,                      -- NC
               --  UART pins
               rxd_i        => uart_rxd_i,
               txd_o        => uart_txd_o );


  -- - - - - - - - - - - - -
  --  output FIFO
  -- - - - - - - - - - - - -
  uart_tx_pop_w <= not uart_wr_busy_w;

  u_fifo_tx : dbg_bridge_fifo
    generic map ( WIDTH  => 8,
                  DEPTH  => 8,
                  ADDR_W => 3 )
    port map ( clk_i     => clk_i,
               xrst_i     => xrst_i,
               --  in
               push_i     => tx_valid_w,
               data_in_i  => tx_data_w,
               accept_o   => tx_accept_w,
               --  out
               pop_i      => uart_tx_pop_w,
               data_out_o => uart_wr_data_w,
               valid_o    => uart_wr_w );


  -- - - - - - - - - - - - -
  --  input FIFO
  -- - - - - - - - - - - - -
  u_fifo_rx : dbg_bridge_fifo
    generic map ( WIDTH  => 8,
                  DEPTH  => 8,
                  ADDR_W => 3 )
    port map ( clk_i      => clk_i,
               xrst_i     => xrst_i,
               --  in
               push_i     => uart_rd_valid_w,
               data_in_i  => uart_rd_data_w,
               accept_o   => uart_rd_w,
               --  out
               pop_i      => rx_accept_w,
               data_out_o => rx_data_w,
               valid_o    => rx_valid_w );


  -- - - - - - - - - - - - -
  --  states
  -- - - - - - - - - - - - -
  p_c_state : process (state_q, rx_valid_w, rx_data_w, mem_wr_q,
                       len_q, mem_bvalid_i, magic_addr_w, mem_rvalid_i,
                       read_skip_w, tx_accept_w)
  begin
    --
    next_state_w <= state_q;
    --
    case state_q is
      -- - - - - - - - - - -
      --  IDLE:
      -- - - - - - - - - - -
      when STATE_IDLE =>
        if (rx_valid_w = '1') then
          case rx_data_w is
            when REQ_WRITE | REQ_READ =>
              next_state_w <= STATE_LEN;
            when others =>
              null;
          end case;
        end if;
      -- - - - - - - - - - -
      --  STATE_LEN
      -- - - - - - - - - - -
      when STATE_LEN =>
        if (rx_valid_w = '1') then
          next_state_w  <= STATE_ADDR0;
        end if;
       -- - - - - - - - - - -
       --  STATE_ADDR
       -- - - - - - - - - - -
      when STATE_ADDR0 =>
        if (rx_valid_w = '1') then
          next_state_w <= STATE_ADDR1;
        end if;
      when STATE_ADDR1 =>
        if (rx_valid_w = '1') then
          next_state_w  <= STATE_ADDR2;
        end if;
      when STATE_ADDR2 =>
        if (rx_valid_w = '1') then
          next_state_w  <= STATE_ADDR3;
        end if;
      when STATE_ADDR3 =>
        if (rx_valid_w = '1' and mem_wr_q = '1') then
          next_state_w  <= STATE_WRITE;
        elsif (rx_valid_w = '1') then
          next_state_w  <= STATE_READ;
        end if;
      -- - - - - - - - - - -
      --  STATE_WRITE
      -- - - - - - - - - - -
      when STATE_WRITE =>
        if (len_q = X"00" and (mem_bvalid_i = '1' or magic_addr_w = '1')) then
          next_state_w  <= STATE_IDLE;
        else
          next_state_w  <= STATE_WRITE;
        end if;
      -- - - - - - - - - - -
      --  STATE_READ
      -- - - - - - - - - - -
      when STATE_READ =>
        --  data ready
        if (mem_rvalid_i = '1' or magic_addr_w = '1') then
          next_state_w  <= STATE_DATA0;
        end if;
      -- - - - - - - - - - -
      --  STATE_DATA
      -- - - - - - - - - - -
      when STATE_DATA0 =>
        if (read_skip_w = '1') then
          next_state_w  <= STATE_DATA1;
        elsif (tx_accept_w  = '1' and (len_q = X"00")) then
          next_state_w  <= STATE_IDLE;
        elsif (tx_accept_w = '1') then
          next_state_w  <= STATE_DATA1;
        end if;
      when STATE_DATA1 =>
        if (read_skip_w = '1') then
          next_state_w  <= STATE_DATA2;
        elsif (tx_accept_w  = '1' and (len_q = X"00")) then
          next_state_w  <= STATE_IDLE;
        elsif (tx_accept_w = '1') then
          next_state_w  <= STATE_DATA2;
        end if;
      when STATE_DATA2 =>
        if (read_skip_w = '1') then
          next_state_w  <= STATE_DATA3;
        elsif (tx_accept_w  = '1' and (len_q = X"00")) then
          next_state_w  <= STATE_IDLE;
        elsif (tx_accept_w = '1') then
          next_state_w  <= STATE_DATA3;
        end if;
      when STATE_DATA3 =>
        if (tx_accept_w  = '1' and (len_q /=  X"00")) then
          next_state_w  <= STATE_READ;
        elsif (tx_accept_w = '1') then
          next_state_w  <= STATE_IDLE;
        end if;
      when others =>
        null;
    end case;
  end process p_c_state;


  --  state storage
  p_state : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        state_q <= STATE_IDLE;
      else
        state_q <= next_state_w;
      end if;
    end if;
  end process p_state;


  -- - - - - - - - - - - - -
  --  RD/WR to and from UART
  -- - - - - - - - - - - - -

  --  write to UART TX buffer in the following states
  tx_valid_w <= '1' when ( ((state_q = STATE_DATA0) or
                            (state_q = STATE_DATA1) or
                            (state_q = STATE_DATA2) or
                            (state_q = STATE_DATA3)) and read_skip_w = '0' )
                else '0';

  --  accept data in the following states
  rx_accept_w <= '1' when ( (state_q = STATE_IDLE)  or
                            (state_q = STATE_LEN)   or
                            (state_q = STATE_ADDR0) or
                            (state_q = STATE_ADDR1) or
                            (state_q = STATE_ADDR2) or
                            (state_q = STATE_ADDR3) or
                            (state_q = STATE_WRITE and mem_busy_q = '0') )
                 else '0';


  -- - - - - - - - - - - - -
  --  capture length
  -- - - - - - - - - - - - -
  p_len : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        len_q       <= X"00";
      else
        if (state_q = STATE_LEN and rx_valid_w = '1') then
          len_q(7 downto 0)  <= rx_data_w;
        elsif (state_q = STATE_WRITE and rx_valid_w = '1' and
               mem_busy_q = '0') then
          len_q       <= dec1u(len_q);
        elsif (state_q = STATE_READ and
               ((mem_busy_q = '1' and mem_rvalid_i = '1') or
                magic_addr_w = '1')) then
          len_q  <= dec1u(len_q);
        elsif (((state_q = STATE_DATA0) or (state_q = STATE_DATA1) or
                (state_q = STATE_DATA2)) and
               (tx_accept_w = '1' and read_skip_w = '0')) then
          len_q  <= dec1u(len_q);
        end if;
      end if;
    end if;
  end process p_len;


  -- - - - - - - - - - - - -
  --  capture addr
  -- - - - - - - - - - - - -
  p_maddr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_addr_q  <= (mem_addr_q'range => '0');
      else
        if (state_q = STATE_ADDR0 and rx_valid_w = '1') then
          mem_addr_q(31 downto 24) <= rx_data_w;
        elsif (state_q = STATE_ADDR1 and rx_valid_w = '1') then
          mem_addr_q(23 downto 16) <= rx_data_w;
        elsif (state_q = STATE_ADDR2 and rx_valid_w = '1') then
          mem_addr_q(15 downto 8)  <= rx_data_w;
        elsif (state_q = STATE_ADDR3 and rx_valid_w = '1') then
          mem_addr_q(7 downto 0)   <= rx_data_w;
          -- Address increment on every access issued
        elsif (state_q = STATE_WRITE and
               (mem_busy_q = '1' and mem_bvalid_i = '1')) then
          mem_addr_q        <= addu(mem_addr_q, 4);
        elsif (state_q = STATE_READ and
               (mem_busy_q  = '1'and mem_rvalid_i = '1')) then
          mem_addr_q        <= addu(mem_addr_q, 4);
        end if;
      end if;
    end if;
  end process p_maddr;


  -- - - - - - - - - - - - -
  --  data index
  -- - - - - - - - - - - - -
  p_didx : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        data_idx_q <= "00";
      else
        if (state_q = STATE_ADDR3) then
          data_idx_q <= rx_data_w(1 downto 0);
        elsif (state_q = STATE_WRITE and
               rx_valid_w  = '1'and mem_busy_q = '0') then
          data_idx_q <= inc1u(data_idx_q);
        elsif (((state_q = STATE_DATA0) or (state_q = STATE_DATA1) or
                (state_q = STATE_DATA2)) and tx_accept_w = '1' and
               (data_idx_q /= "00")) then
          data_idx_q <= dec1u(data_idx_q);
        end if;
      end if;
    end if;
  end process p_didx;


  read_skip_w <= not_zero(data_idx_q);



  -- - - - - - - - - - - - -
  --  data sample
  -- - - - - - - - - - - - -
  p_data : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        data_q <= X"00000000";
      --  write to memory
      else
        if (state_q = STATE_WRITE and rx_valid_w = '1' and
            mem_busy_q = '0') then
          case data_idx_q is
            when "00" => data_q(7 downto 0)   <= rx_data_w;
            when "01" => data_q(15 downto 8)  <= rx_data_w;
            when "10" => data_q(23 downto 16) <= rx_data_w;
            when "11" => data_q(31 downto 24) <= rx_data_w;
            when others =>
              null;
          end case;
        --  read from GPIO input?
        elsif (state_q = STATE_READ and mem_addr_q = GPIO_ADDRESS) then
          data_q <= gpio_in_w;
        --  read from status register?
        elsif (state_q = STATE_READ and mem_addr_q = STS_ADDRESS) then
          data_q <= (X"CAFE" & (14 downto 0 => '0') & mem_busy_q);
        --  read from memory
        elsif (state_q = STATE_READ and mem_rvalid_i = '1') then
          data_q <= mem_rdata_i;
        --  shift data out (read response -> UART)
        elsif (((state_q = STATE_DATA0) or (state_q = STATE_DATA1) or
                (state_q = STATE_DATA2)) and
               (tx_accept_w = '1' or read_skip_w = '1')) then
          data_q <= (X"00" & data_q(31 downto 8));
        end if;
      end if;
    end if;
  end process p_data;


  tx_data_w  <= data_q(7 downto 0);

  mem_wdata_o <= data_q;


  -- - - - - - - - - - - - -
  --  AXI: write request
  -- - - - - - - - - - - - -
  p_c_axi_wreq : process (mem_awvalid_o_w, mem_awready_i, mem_awvalid_q,
                         state_q, rx_valid_w, data_idx_q, len_q, magic_addr_w,
                         mem_wvalid_o_w, mem_wready_i, mem_wvalid_q)
  begin
    mem_awvalid_w <= '0';
    mem_wvalid_w  <= '0';
    --  hold
    if (mem_awvalid_o_w = '1' and  mem_awready_i = '0') then
      mem_awvalid_w <= mem_awvalid_q;
    elsif (mem_awvalid_o_w = '1') then
      mem_awvalid_w <= '0';
      --  every 4th byte, issue bus access
    elsif (state_q = STATE_WRITE and rx_valid_w = '1' and
           (data_idx_q = "11" or unsigned(len_q) = 1)) then
      mem_awvalid_w <= not magic_addr_w;
    end if;
    --  hold
    if (mem_wvalid_o_w = '1' and mem_wready_i = '0') then
      mem_wvalid_w <= mem_wvalid_q;
    elsif (mem_wvalid_o_w = '1') then
      mem_wvalid_w <= '0';
      --  every 4th byte, issue bus access
    elsif (state_q = STATE_WRITE and rx_valid_w = '1' and
           (data_idx_q = "11" or unsigned(len_q) = 1)) then
      mem_wvalid_w <= not magic_addr_w;
    end if;
  end process p_c_axi_wreq;


  p_mvalid : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_awvalid_q <= '0';
        mem_wvalid_q  <= '0';
      else
        mem_awvalid_q <= mem_awvalid_w;
        mem_wvalid_q  <= mem_wvalid_w;
      end if;
    end if;
  end process p_mvalid;


  mem_awvalid_o_w <= mem_awvalid_q;
  mem_awvalid_o   <= mem_awvalid_o_w;

  mem_wvalid_o_w  <= mem_wvalid_q;
  mem_wvalid_o    <= mem_wvalid_o_w;

  mem_awaddr_o    <= (mem_addr_q(31 downto 2) & "00");

  mem_bready_o    <= '1';


  -- - - - - - - - - - - - -
  --  AXI: Read Request
  -- - - - - - - - - - - - -
  p_c_axi_rreq : process (mem_arvalid_o_w, mem_arready_i, mem_arvalid_q,
                          mem_busy_q, state_q, magic_addr_w)
  begin
    mem_arvalid_w <= '0';
    --  hold
    if (mem_arvalid_o_w = '1' and mem_arready_i = '0') then
      mem_arvalid_w <= mem_arvalid_q;
    elsif (mem_arvalid_o_w = '1') then
      mem_arvalid_w <= '0';
    elsif (state_q = STATE_READ and mem_busy_q = '0') then
      mem_arvalid_w <= not magic_addr_w;
    end if;
  end process p_c_axi_rreq;


  p_arvalid : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_arvalid_q <= '0';
      else
        mem_arvalid_q <= mem_arvalid_w;
      end if;
    end if;
  end process p_arvalid;


  mem_arvalid_o_w <= mem_arvalid_q;
  mem_arvalid_o   <= mem_arvalid_o_w;

  mem_araddr_o    <= (mem_addr_q(31 downto 2) & "00");

  mem_rready_o    <= '1';


  -- - - - - - - - - - - - -
  --  write mask
  -- - - - - - - - - - - - -
  p_c_msel : process (data_idx_q, mem_addr_q)
    variable mem_sel_v : slv_t(mem_sel_w'range);
  begin
    mem_sel_v := "1111";
    --
    case data_idx_q is
      when "00" => mem_sel_v := "0001";
      when "01" => mem_sel_v := "0011";
      when "10" => mem_sel_v := "0111";
      when "11" => mem_sel_v := "1111";
      when others => null;
    end case;
    --
    case mem_addr_q(1 downto 0) is
      when "00" => mem_sel_v := mem_sel_v and "1111";
      when "01" => mem_sel_v := mem_sel_v and "1110";
      when "10" => mem_sel_v := mem_sel_v and "1100";
      when "11" => mem_sel_v := mem_sel_v and "1000";
      when others => null;
    end case;
    mem_sel_w <= mem_sel_v;
  end process p_c_msel;


  p_msel : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_sel_q    <= X"0";
      --  idle - reset for read requests
      else
        if (state_q = STATE_IDLE) then
          mem_sel_q   <= "1111";
        --  every 4th byte, issue bus access
        elsif (state_q = STATE_WRITE and rx_valid_w = '1' and
               (data_idx_q = "11" or unsigned(len_q) = 1)) then
          mem_sel_q   <= mem_sel_w;
        end if;
      end if;
    end if;
  end process p_msel;


  mem_wstrb_o  <= mem_sel_q;


  -- - - - - - - - - - - - -
  --  write enable
  -- - - - - - - - - - - - -
  p_mwr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_wr_q    <= '0';
      else
        if (state_q = STATE_IDLE and rx_valid_w = '1') then
          if (rx_data_w = REQ_WRITE) then
            mem_wr_q <= '1';
          else
            mem_wr_q <= '0';
          end if;
        end if;
      end if;
    end if;
  end process p_mwr;


  -- - - - - - - - - - - - -
  --  access in progress
  -- - - - - - - - - - - - -
  p_mbusy : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        mem_busy_q <= '0';
      else
        if (mem_arvalid_o_w = '1' or mem_awvalid_o_w = '1') then
          mem_busy_q <= '1';
        elsif (mem_bvalid_i = '1' or mem_rvalid_i = '1') then
          mem_busy_q <= '0';
        end if;
      end if;
    end if;
  end process p_mbusy;


end rtl;
