-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity rv_soc is
  generic ( CLK_FREQ    : natural := 50000000;           -- default 50M
            UART_SPEED  : natural :=  1000000;           -- default 1M
            NR_GPIOS    : integer range 1 to 32 := 8 );
  port ( clk_i           : in  std_logic;
         xrst_i          : in  std_logic;
         --
         xmem_awready_i  : in  std_logic;                -- xmem
         xmem_wready_i   : in  std_logic;
         xmem_bvalid_i   : in  std_logic;
         xmem_bresp_i    : in  slv_01_00;
         xmem_bid_i      : in  slv_03_00;
         xmem_arready_i  : in  std_logic;
         xmem_rvalid_i   : in  std_logic;
         xmem_rdata_i    : in  slv_31_00;
         xmem_rresp_i    : in  slv_01_00;
         xmem_rid_i      : in  slv_03_00;
         xmem_rlast_i    : in  std_logic;
         xmem_awvalid_o  : out std_logic;
         xmem_awaddr_o   : out slv_31_00;
         xmem_awid_o     : out slv_03_00;
         xmem_awlen_o    : out slv_07_00;
         xmem_awburst_o  : out slv_01_00;
         xmem_wvalid_o   : out std_logic;
         xmem_wdata_o    : out slv_31_00;
         xmem_wstrb_o    : out slv_03_00;
         xmem_wlast_o    : out std_logic;
         xmem_bready_o   : out std_logic;
         xmem_arvalid_o  : out std_logic;
         xmem_araddr_o   : out slv_31_00;
         xmem_arid_o     : out slv_03_00;
         xmem_arlen_o    : out slv_07_00;
         xmem_arburst_o  : out slv_01_00;
         xmem_rready_o   : out std_logic;
         --
         uart_txd_i      : in  std_logic;                 -- io pins
         uart_rxd_o      : out std_logic;
         uart2_txd_i     : in  std_logic;
         uart2_rxd_o     : out std_logic;
         --
         spi_miso_i      : in  std_logic;
         spi_clk_o       : out std_logic;
         spi_cs_o        : out std_logic;
         spi_mosi_o      : out std_logic;
         --
         gpio_i          : in  slv_t(NR_GPIOS-1 downto 0);
         gpio_oena_o     : out slv_t(NR_GPIOS-1 downto 0);
         gpio_o          : out slv_t(NR_GPIOS-1 downto 0)
         );
end entity rv_soc;


architecture rtl of rv_soc is

  component rv_core
    generic ( BOOT_ADDR      : slv_31_00 := X"80000000";
              CPU_ID         : slv_31_00 := X"00000000";
              ICACHE_AXI_ID  : slv_03_00 := X"8";
              DCACHE_AXI_ID  : slv_03_00 := X"4";
              CACHE_ADDR_L   : slv_31_00 := X"80000000";
              CACHE_ADDR_H   : slv_31_00 := X"8FFFFFFF" );
    port ( clk_i            : in  std_logic;
           xrst_i           : in  std_logic;             -- syncr reset for cpu
           xrst_cpu_i       : in  std_logic;
           --
           axi_i_awready_i  : in  std_logic;             -- I-bus AXI 11 + 16
           axi_i_wready_i   : in  std_logic;
           axi_i_bvalid_i   : in  std_logic;
           axi_i_bresp_i    : in  slv_01_00;
           axi_i_bid_i      : in  slv_03_00;
           axi_i_arready_i  : in  std_logic;
           axi_i_rvalid_i   : in  std_logic;
           axi_i_rdata_i    : in  slv_31_00;
           axi_i_rresp_i    : in  slv_01_00;
           axi_i_rid_i      : in  slv_03_00;
           axi_i_rlast_i    : in  std_logic;
           axi_i_awvalid_o  : out std_logic;
           axi_i_awaddr_o   : out slv_31_00;
           axi_i_awid_o     : out slv_03_00;
           axi_i_awlen_o    : out slv_07_00;
           axi_i_awburst_o  : out slv_01_00;
           axi_i_wvalid_o   : out std_logic;
           axi_i_wdata_o    : out slv_31_00;
           axi_i_wstrb_o    : out slv_03_00;
           axi_i_wlast_o    : out std_logic;
           axi_i_bready_o   : out std_logic;
           axi_i_arvalid_o  : out std_logic;
           axi_i_araddr_o   : out slv_31_00;
           axi_i_arid_o     : out slv_03_00;
           axi_i_arlen_o    : out slv_07_00;
           axi_i_arburst_o  : out slv_01_00;
           axi_i_rready_o   : out std_logic;
           --
           axi_d_awready_i  : in  std_logic;             -- D-bus AXI 11 + 16
           axi_d_wready_i   : in  std_logic;
           axi_d_bvalid_i   : in  std_logic;
           axi_d_bresp_i    : in  slv_01_00;
           axi_d_bid_i      : in  slv_03_00;
           axi_d_arready_i  : in  std_logic;
           axi_d_rvalid_i   : in  std_logic;
           axi_d_rdata_i    : in  slv_31_00;
           axi_d_rresp_i    : in  slv_01_00;
           axi_d_rid_i      : in  slv_03_00;
           axi_d_rlast_i    : in  std_logic;
           axi_d_awvalid_o  : out std_logic;
           axi_d_awaddr_o   : out slv_31_00;
           axi_d_awid_o     : out slv_03_00;
           axi_d_awlen_o    : out slv_07_00;
           axi_d_awburst_o  : out slv_01_00;
           axi_d_wvalid_o   : out std_logic;
           axi_d_wdata_o    : out slv_31_00;
           axi_d_wstrb_o    : out slv_03_00;
           axi_d_wlast_o    : out std_logic;
           axi_d_bready_o   : out std_logic;
           axi_d_arvalid_o  : out std_logic;
           axi_d_araddr_o   : out slv_31_00;
           axi_d_arid_o     : out slv_03_00;
           axi_d_arlen_o    : out slv_07_00;
           axi_d_arburst_o  : out slv_01_00;
           axi_d_rready_o   : out std_logic;
           --  vsbus for perips
           pp_irq_i         : in  std_logic;             -- perip interrupt
           pp_addr_o        : out slv_31_00;
           pp_rd_o          : out std_logic;
           pp_rd_data_i     : in  slv_31_00;
           pp_wr_o          : out slv_03_00;
           pp_wr_data_o     : out slv_31_00 );
  end component;


  component vsperips
    generic ( CLK_FREQ   : natural := 50000000;          -- default 50M
              UART_SPEED : natural :=  1000000;          -- default 1M
              NR_GPIOS   : integer range 1 to 32 := 8 );
    port ( clk_i         : in  std_logic;
           xrst_i        : in  std_logic;                -- syncr reset
           --  vsbus for peripherals
           pp_addr_i     : in  slv_31_00;
           pp_rd_i       : in  std_logic;
           pp_wr_i       : in  slv_03_00;
           pp_wr_data_i  : in  slv_31_00;
           pp_rd_data_o  : out slv_31_00;
           pp_irq_o      : out std_logic;                -- pp interrupt
           --
           --  peripheral ios to pads
           rx_i          : in  std_logic;
           tx_o          : out std_logic;
           rx2_i         : in  std_logic;
           tx2_o         : out std_logic;
           --
           spi_miso_i    : in  std_logic;
           spi_clk_o     : out std_logic;
           spi_cs_o      : out std_logic;
           spi_mosi_o    : out std_logic;
           --
           gpio_i        : in  slv_t(NR_GPIOS-1 downto 0);
           gpio_oena_o   : out slv_t(NR_GPIOS-1 downto 0);
           gpio_o        : out slv_t(NR_GPIOS-1 downto 0) );
  end component;


  component axi_mem_if
    port ( clk_i             : in  std_logic;
           xrst_i            : in  std_logic;            -- syncr reset
           --
           inport_awvalid_i  : in  std_logic;            -- dbg uart
           inport_awaddr_i   : in  slv_31_00;
           inport_awid_i     : in  slv_03_00;
           inport_awlen_i    : in  slv_07_00;
           inport_awburst_i  : in  slv_01_00;
           inport_wvalid_i   : in  std_logic;
           inport_wdata_i    : in  slv_31_00;
           inport_wstrb_i    : in  slv_03_00;
           inport_wlast_i    : in  std_logic;
           inport_bready_i   : in  std_logic;
           inport_arvalid_i  : in  std_logic;
           inport_araddr_i   : in  slv_31_00;
           inport_arid_i     : in  slv_03_00;
           inport_arlen_i    : in  slv_07_00;
           inport_arburst_i  : in  slv_01_00;
           inport_rready_i   : in  std_logic;
           inport_awready_o  : out std_logic;
           inport_wready_o   : out std_logic;
           inport_bvalid_o   : out std_logic;
           inport_bresp_o    : out slv_01_00;
           inport_bid_o      : out slv_03_00;
           inport_arready_o  : out std_logic;
           inport_rvalid_o   : out std_logic;
           inport_rdata_o    : out slv_31_00;
           inport_rresp_o    : out slv_01_00;
           inport_rid_o      : out slv_03_00;
           inport_rlast_o    : out std_logic;
           --
           cpu_i_awvalid_i   : in  std_logic;            -- I-bus
           cpu_i_awaddr_i    : in  slv_31_00;
           cpu_i_awid_i      : in  slv_03_00;
           cpu_i_awlen_i     : in  slv_07_00;
           cpu_i_awburst_i   : in  slv_01_00;
           cpu_i_wvalid_i    : in  std_logic;
           cpu_i_wdata_i     : in  slv_31_00;
           cpu_i_wstrb_i     : in  slv_03_00;
           cpu_i_wlast_i     : in  std_logic;
           cpu_i_bready_i    : in  std_logic;
           cpu_i_arvalid_i   : in  std_logic;
           cpu_i_araddr_i    : in  slv_31_00;
           cpu_i_arid_i      : in  slv_03_00;
           cpu_i_arlen_i     : in  slv_07_00;
           cpu_i_arburst_i   : in  slv_01_00;
           cpu_i_rready_i    : in  std_logic;
           cpu_i_awready_o   : out std_logic;
           cpu_i_wready_o    : out std_logic;
           cpu_i_bvalid_o    : out std_logic;
           cpu_i_bresp_o     : out slv_01_00;
           cpu_i_bid_o       : out slv_03_00;
           cpu_i_arready_o   : out std_logic;
           cpu_i_rvalid_o    : out std_logic;
           cpu_i_rdata_o     : out slv_31_00;
           cpu_i_rresp_o     : out slv_01_00;
           cpu_i_rid_o       : out slv_03_00;
           cpu_i_rlast_o     : out std_logic;
           --
           cpu_d_awvalid_i   : in  std_logic;            -- D-bus
           cpu_d_awaddr_i    : in  slv_31_00;
           cpu_d_awid_i      : in  slv_03_00;
           cpu_d_awlen_i     : in  slv_07_00;
           cpu_d_awburst_i   : in  slv_01_00;
           cpu_d_wvalid_i    : in  std_logic;
           cpu_d_wdata_i     : in  slv_31_00;
           cpu_d_wstrb_i     : in  slv_03_00;
           cpu_d_wlast_i     : in  std_logic;
           cpu_d_bready_i    : in  std_logic;
           cpu_d_arvalid_i   : in  std_logic;
           cpu_d_araddr_i    : in  slv_31_00;
           cpu_d_arid_i      : in  slv_03_00;
           cpu_d_arlen_i     : in  slv_07_00;
           cpu_d_arburst_i   : in  slv_01_00;
           cpu_d_rready_i    : in  std_logic;
           cpu_d_awready_o   : out std_logic;
           cpu_d_wready_o    : out std_logic;
           cpu_d_bvalid_o    : out std_logic;
           cpu_d_bresp_o     : out slv_01_00;
           cpu_d_bid_o       : out slv_03_00;
           cpu_d_arready_o   : out std_logic;
           cpu_d_rvalid_o    : out std_logic;
           cpu_d_rdata_o     : out slv_31_00;
           cpu_d_rresp_o     : out slv_01_00;
           cpu_d_rid_o       : out slv_03_00;
           cpu_d_rlast_o     : out std_logic;
           --
           mem_awready_i     : in  std_logic;            -- zmem
           mem_wready_i      : in  std_logic;
           mem_bvalid_i      : in  std_logic;
           mem_bresp_i       : in  slv_01_00;
           mem_bid_i         : in  slv_03_00;
           mem_arready_i     : in  std_logic;
           mem_rvalid_i      : in  std_logic;
           mem_rdata_i       : in  slv_31_00;
           mem_rresp_i       : in  slv_01_00;
           mem_rid_i         : in  slv_03_00;
           mem_rlast_i       : in  std_logic;
           mem_awvalid_o     : out std_logic;
           mem_awaddr_o      : out slv_31_00;
           mem_awid_o        : out slv_03_00;
           mem_awlen_o       : out slv_07_00;
           mem_awburst_o     : out slv_01_00;
           mem_wvalid_o      : out std_logic;
           mem_wdata_o       : out slv_31_00;
           mem_wstrb_o       : out slv_03_00;
           mem_wlast_o       : out std_logic;
           mem_bready_o      : out std_logic;
           mem_arvalid_o     : out std_logic;
           mem_araddr_o      : out slv_31_00;
           mem_arid_o        : out slv_03_00;
           mem_arlen_o       : out slv_07_00;
           mem_arburst_o     : out slv_01_00;
           mem_rready_o      : out std_logic );
  end component;


  component dbg_bridge
    generic ( CLK_FREQ    : natural := 50000000;         -- default 50M
              UART_SPEED  : natural :=  1000000 );       -- default 1M
    port ( clk_i          : in  std_logic;
           xrst_i         : in  std_logic;
           --  AXI bus to DDR / TCM ram if
           mem_awready_i  : in  std_logic;
           mem_wready_i   : in  std_logic;
           mem_bvalid_i   : in  std_logic;
           mem_arready_i  : in  std_logic;
           mem_rvalid_i   : in  std_logic;
           mem_rdata_i    : in  slv_31_00;
           mem_awvalid_o  : out std_logic;
           mem_awaddr_o   : out slv_31_00;
           mem_wvalid_o   : out std_logic;
           mem_wdata_o    : out slv_31_00;
           mem_wstrb_o    : out slv_03_00;
           mem_bready_o   : out std_logic;
           mem_arvalid_o  : out std_logic;
           mem_araddr_o   : out slv_31_00;
           mem_rready_o   : out std_logic;
           --  pp io
           uart_rxd_i     : in  std_logic;
           uart_txd_o     : out std_logic;
           rv_core_ena_o  : out std_logic );             -- cpu / cache enable
  end component;

  
  signal axi4_i_rvalid_w   : std_logic;
  signal axi4_i_awaddr_w   : slv_31_00;
  signal axi4_i_arvalid_w  : std_logic;
  signal axi4_i_wstrb_w    : slv_03_00;
  signal axi4_i_awlen_w    : slv_07_00;
  signal axi4_i_wlast_w    : std_logic;
  signal axi4_i_bvalid_w   : std_logic;
  signal axi4_i_awvalid_w  : std_logic;
  signal axi4_i_rlast_w    : std_logic;
  signal axi4_i_wdata_w    : slv_31_00;
  signal axi4_i_bready_w   : std_logic;
  signal axi4_i_wready_w   : std_logic;
  signal axi4_i_bid_w      : slv_03_00;
  signal axi4_i_wvalid_w   : std_logic;
  signal axi4_i_rready_w   : std_logic;
  signal axi4_i_arready_w  : std_logic;
  signal axi4_i_bresp_w    : slv_01_00;
  signal axi4_i_arid_w     : slv_03_00;
  signal axi4_i_arlen_w    : slv_07_00;
  signal axi4_i_awid_w     : slv_03_00;
  signal axi4_i_araddr_w   : slv_31_00;
  signal axi4_i_rresp_w    : slv_01_00;
  signal axi4_i_arburst_w  : slv_01_00;
  signal axi4_i_awready_w  : std_logic;
  signal axi4_i_awburst_w  : slv_01_00;
  signal axi4_i_rid_w      : slv_03_00;
  signal axi4_i_rdata_w    : slv_31_00;

  signal axi4_d_awid_w     : slv_03_00;
  signal axi4_d_rdata_w    : slv_31_00;
  signal axi4_d_awaddr_w   : slv_31_00;
  signal axi4_d_awvalid_w  : std_logic;
  signal axi4_d_arlen_w    : slv_07_00;
  signal axi4_d_rready_w   : std_logic;
  signal axi4_d_awburst_w  : slv_01_00;
  signal axi4_d_bresp_w    : slv_01_00;
  signal axi4_d_bvalid_w   : std_logic;
  signal axi4_d_awready_w  : std_logic;
  signal axi4_d_bid_w      : slv_03_00;
  signal axi4_d_rresp_w    : slv_01_00;
  signal axi4_d_bready_w   : std_logic;
  signal axi4_d_wdata_w    : slv_31_00;
  signal axi4_d_wstrb_w    : slv_03_00;
  signal axi4_d_arburst_w  : slv_01_00;
  signal axi4_d_araddr_w   : slv_31_00;
  signal axi4_d_rvalid_w   : std_logic;
  signal axi4_d_rid_w      : slv_03_00;
  signal axi4_d_arid_w     : slv_03_00;
  signal axi4_d_arvalid_w  : std_logic;
  signal axi4_d_awlen_w    : slv_07_00;
  signal axi4_d_arready_w  : std_logic;
  signal axi4_d_wready_w   : std_logic;
  signal axi4_d_rlast_w    : std_logic;
  signal axi4_d_wvalid_w   : std_logic;
  signal axi4_d_wlast_w    : std_logic;

  signal pp_irq_w          : std_logic;                  -- perip interrupt
  signal pp_addr_w         : slv_31_00;
  signal pp_rd_w           : std_logic;
  signal pp_rd_data_w      : slv_31_00;
  signal pp_wr_w           : slv_03_00;
  signal pp_wr_data_w      : slv_31_00;


  signal axi_awready_w     : std_logic;                  -- dbg uart
  signal axi_wready_w      : std_logic;
  signal axi_bvalid_w      : std_logic;
  --
  signal axi_awvalid_w     : std_logic;
  signal axi_awaddr_w      : slv_31_00;
  signal axi_wvalid_w      : std_logic;
  signal axi_wdata_w       : slv_31_00;
  signal axi_wstrb_w       : slv_03_00;
  signal axi_bready_w      : std_logic;
  --
  signal axi_arready_w     : std_logic;
  signal axi_rvalid_w      : std_logic;
  signal axi_rdata_w       : slv_31_00;
  --
  signal axi_arvalid_w     : std_logic;
  signal axi_araddr_w      : slv_31_00;
  signal axi_rready_w      : std_logic;


  signal xrst_cpu_w        : std_logic;
  signal rv_core_ena_w     : std_logic;

  signal uart_txd_w        : std_logic;
  signal dbg_txd_w         : std_logic;
  signal txd_q             : std_logic;

begin


  u_core : rv_core
    port map ( clk_i            => clk_i,
               xrst_i           => xrst_i,
               xrst_cpu_i       => xrst_cpu_w,           -- NOTE: cpu reset
               --
               axi_i_awready_i  => axi4_i_awready_w,     -- I-bus AXI 11+16
               axi_i_wready_i   => axi4_i_wready_w,
               axi_i_bvalid_i   => axi4_i_bvalid_w,
               axi_i_bresp_i    => axi4_i_bresp_w,
               axi_i_bid_i      => axi4_i_bid_w,
               axi_i_arready_i  => axi4_i_arready_w,
               axi_i_rvalid_i   => axi4_i_rvalid_w,
               axi_i_rdata_i    => axi4_i_rdata_w,
               axi_i_rresp_i    => axi4_i_rresp_w,
               axi_i_rid_i      => axi4_i_rid_w,
               axi_i_rlast_i    => axi4_i_rlast_w,
               axi_i_awvalid_o  => axi4_i_awvalid_w,
               axi_i_awaddr_o   => axi4_i_awaddr_w,
               axi_i_awid_o     => axi4_i_awid_w,
               axi_i_awlen_o    => axi4_i_awlen_w,
               axi_i_awburst_o  => axi4_i_awburst_w,
               axi_i_wvalid_o   => axi4_i_wvalid_w,
               axi_i_wdata_o    => axi4_i_wdata_w,
               axi_i_wstrb_o    => axi4_i_wstrb_w,
               axi_i_wlast_o    => axi4_i_wlast_w,
               axi_i_bready_o   => axi4_i_bready_w,
               axi_i_arvalid_o  => axi4_i_arvalid_w,
               axi_i_araddr_o   => axi4_i_araddr_w,
               axi_i_arid_o     => axi4_i_arid_w,
               axi_i_arlen_o    => axi4_i_arlen_w,
               axi_i_arburst_o  => axi4_i_arburst_w,
               axi_i_rready_o   => axi4_i_rready_w,
               --
               axi_d_awready_i  => axi4_d_awready_w,     -- D-bus AXI 11+16
               axi_d_wready_i   => axi4_d_wready_w,
               axi_d_bvalid_i   => axi4_d_bvalid_w,
               axi_d_bresp_i    => axi4_d_bresp_w,
               axi_d_bid_i      => axi4_d_bid_w,
               axi_d_arready_i  => axi4_d_arready_w,
               axi_d_rvalid_i   => axi4_d_rvalid_w,
               axi_d_rdata_i    => axi4_d_rdata_w,
               axi_d_rresp_i    => axi4_d_rresp_w,
               axi_d_rid_i      => axi4_d_rid_w,
               axi_d_rlast_i    => axi4_d_rlast_w,
               axi_d_awvalid_o  => axi4_d_awvalid_w,
               axi_d_awaddr_o   => axi4_d_awaddr_w,
               axi_d_awid_o     => axi4_d_awid_w,
               axi_d_awlen_o    => axi4_d_awlen_w,
               axi_d_awburst_o  => axi4_d_awburst_w,
               axi_d_wvalid_o   => axi4_d_wvalid_w,
               axi_d_wdata_o    => axi4_d_wdata_w,
               axi_d_wstrb_o    => axi4_d_wstrb_w,
               axi_d_wlast_o    => axi4_d_wlast_w,
               axi_d_bready_o   => axi4_d_bready_w,
               axi_d_arvalid_o  => axi4_d_arvalid_w,
               axi_d_araddr_o   => axi4_d_araddr_w,
               axi_d_arid_o     => axi4_d_arid_w,
               axi_d_arlen_o    => axi4_d_arlen_w,
               axi_d_arburst_o  => axi4_d_arburst_w,
               axi_d_rready_o   => axi4_d_rready_w,
               --  vsbus for peripherals
               pp_irq_i         => pp_irq_w,             -- perip interrupt
               pp_addr_o        => pp_addr_w,
               pp_rd_o          => pp_rd_w,
               pp_rd_data_i     => pp_rd_data_w,
               pp_wr_o          => pp_wr_w,
               pp_wr_data_o     => pp_wr_data_w );



  u_vsperips : vsperips
    generic map ( CLK_FREQ    => CLK_FREQ,               -- default 50M
                  UART_SPEED  => UART_SPEED,             -- default 1M
                  NR_GPIOS    => NR_GPIOS )
    port map ( clk_i         => clk_i,
               xrst_i        => xrst_i,                  -- syncr reset
               --  vsbus for peripherals
               pp_addr_i     => pp_addr_w,
               pp_rd_i       => pp_rd_w,
               pp_wr_i       => pp_wr_w,
               pp_wr_data_i  => pp_wr_data_w,
               pp_rd_data_o  => pp_rd_data_w,
               pp_irq_o      => pp_irq_w,                -- pp intr
               --
               --  perip IO
               rx_i          => uart_txd_i,              -- in
               tx_o          => uart_txd_w,              -- out
               rx2_i         => uart2_txd_i,             -- in
               tx2_o         => uart2_rxd_o,             -- out
               --
               gpio_i        => gpio_i,
               gpio_oena_o   => gpio_oena_o,
               gpio_o        => gpio_o,
               --
               spi_miso_i    => spi_miso_i,
               spi_clk_o     => spi_clk_o,
               spi_cs_o      => spi_cs_o,
               spi_mosi_o    => spi_mosi_o );



  u_axi_mem_if : axi_mem_if
    port map ( clk_i             => clk_i,
               xrst_i            => xrst_i,
               --
               inport_awready_o  => axi_awready_w,
               inport_wready_o   => axi_wready_w,
               inport_bvalid_o   => axi_bvalid_w,
               inport_bresp_o    => OPEN,                -- !
               inport_bid_o      => OPEN,                -- !
               --
               inport_awvalid_i  => axi_awvalid_w,       -- dbg uart
               inport_awaddr_i   => axi_awaddr_w,
               inport_awid_i     => X"0",                -- !
               inport_awlen_i    => X"00",               -- !
               inport_awburst_i  => "01",                -- !
               inport_wvalid_i   => axi_wvalid_w,
               inport_wdata_i    => axi_wdata_w,
               inport_wstrb_i    => axi_wstrb_w,
               inport_wlast_i    => '1',                 -- !
               inport_bready_i   => axi_bready_w,
               --
               inport_arready_o  => axi_arready_w,
               inport_rvalid_o   => axi_rvalid_w,
               inport_rdata_o    => axi_rdata_w,
               inport_rresp_o    => OPEN,                -- !
               inport_rid_o      => OPEN,                -- !
               inport_rlast_o    => OPEN,                -- !
               --
               inport_arvalid_i  => axi_arvalid_w,
               inport_araddr_i   => axi_araddr_w,
               inport_arid_i     => X"0",                -- !
               inport_arlen_i    => X"00",               -- !
               inport_arburst_i  => "01",                -- !
               inport_rready_i   => axi_rready_w,
               --
               --  cpu i-bus
               cpu_i_awvalid_i   => axi4_i_awvalid_w,    -- I-BUS
               cpu_i_awaddr_i    => axi4_i_awaddr_w,
               cpu_i_awid_i      => axi4_i_awid_w,
               cpu_i_awlen_i     => axi4_i_awlen_w,
               cpu_i_awburst_i   => axi4_i_awburst_w,
               cpu_i_wvalid_i    => axi4_i_wvalid_w,
               cpu_i_wdata_i     => axi4_i_wdata_w,
               cpu_i_wstrb_i     => axi4_i_wstrb_w,
               cpu_i_wlast_i     => axi4_i_wlast_w,
               cpu_i_bready_i    => axi4_i_bready_w,
               cpu_i_arvalid_i   => axi4_i_arvalid_w,
               cpu_i_araddr_i    => axi4_i_araddr_w,
               cpu_i_arid_i      => axi4_i_arid_w,
               cpu_i_arlen_i     => axi4_i_arlen_w,
               cpu_i_arburst_i   => axi4_i_arburst_w,
               cpu_i_rready_i    => axi4_i_rready_w,
               cpu_i_awready_o   => axi4_i_awready_w,
               cpu_i_wready_o    => axi4_i_wready_w,
               cpu_i_bvalid_o    => axi4_i_bvalid_w,
               cpu_i_bresp_o     => axi4_i_bresp_w,
               cpu_i_bid_o       => axi4_i_bid_w,
               cpu_i_arready_o   => axi4_i_arready_w,
               cpu_i_rvalid_o    => axi4_i_rvalid_w,
               cpu_i_rdata_o     => axi4_i_rdata_w,
               cpu_i_rresp_o     => axi4_i_rresp_w,
               cpu_i_rid_o       => axi4_i_rid_w,
               cpu_i_rlast_o     => axi4_i_rlast_w,
               --  cpu d-bus
               cpu_d_awvalid_i   => axi4_d_awvalid_w,    -- D-BUS
               cpu_d_awaddr_i    => axi4_d_awaddr_w,
               cpu_d_awid_i      => axi4_d_awid_w,
               cpu_d_awlen_i     => axi4_d_awlen_w,
               cpu_d_awburst_i   => axi4_d_awburst_w,
               cpu_d_wvalid_i    => axi4_d_wvalid_w,
               cpu_d_wdata_i     => axi4_d_wdata_w,
               cpu_d_wstrb_i     => axi4_d_wstrb_w,
               cpu_d_wlast_i     => axi4_d_wlast_w,
               cpu_d_bready_i    => axi4_d_bready_w,
               cpu_d_arvalid_i   => axi4_d_arvalid_w,
               cpu_d_araddr_i    => axi4_d_araddr_w,
               cpu_d_arid_i      => axi4_d_arid_w,
               cpu_d_arlen_i     => axi4_d_arlen_w,
               cpu_d_arburst_i   => axi4_d_arburst_w,
               cpu_d_rready_i    => axi4_d_rready_w,
               cpu_d_awready_o   => axi4_d_awready_w,
               cpu_d_wready_o    => axi4_d_wready_w,
               cpu_d_bvalid_o    => axi4_d_bvalid_w,
               cpu_d_bresp_o     => axi4_d_bresp_w,
               cpu_d_bid_o       => axi4_d_bid_w,
               cpu_d_arready_o   => axi4_d_arready_w,
               cpu_d_rvalid_o    => axi4_d_rvalid_w,
               cpu_d_rdata_o     => axi4_d_rdata_w,
               cpu_d_rresp_o     => axi4_d_rresp_w,
               cpu_d_rid_o       => axi4_d_rid_w,
               cpu_d_rlast_o     => axi4_d_rlast_w,
               --
               mem_awready_i     => xmem_awready_i,      -- xmem
               mem_wready_i      => xmem_wready_i,
               mem_bvalid_i      => xmem_bvalid_i,
               mem_bresp_i       => xmem_bresp_i,
               mem_bid_i         => xmem_bid_i,
               mem_arready_i     => xmem_arready_i,
               mem_rvalid_i      => xmem_rvalid_i,
               mem_rdata_i       => xmem_rdata_i,
               mem_rresp_i       => xmem_rresp_i,
               mem_rid_i         => xmem_rid_i,
               mem_rlast_i       => xmem_rlast_i,
               mem_awvalid_o     => xmem_awvalid_o,
               mem_awaddr_o      => xmem_awaddr_o,
               mem_awid_o        => xmem_awid_o,
               mem_awlen_o       => xmem_awlen_o,
               mem_awburst_o     => xmem_awburst_o,
               mem_wvalid_o      => xmem_wvalid_o,
               mem_wdata_o       => xmem_wdata_o,
               mem_wstrb_o       => xmem_wstrb_o,
               mem_wlast_o       => xmem_wlast_o,
               mem_bready_o      => xmem_bready_o,
               mem_arvalid_o     => xmem_arvalid_o,
               mem_araddr_o      => xmem_araddr_o,
               mem_arid_o        => xmem_arid_o,
               mem_arlen_o       => xmem_arlen_o,
               mem_arburst_o     => xmem_arburst_o,
               mem_rready_o      => xmem_rready_o );


  u_dbg_bridge : dbg_bridge
    generic map ( CLK_FREQ    => CLK_FREQ,                -- default 50M
                  UART_SPEED  => UART_SPEED )             -- default 1M
    port map ( clk_i          => clk_i,
               xrst_i         => xrst_i,
               --  AXI bus to DDR / TCM ram if
               mem_awready_i  => axi_awready_w,
               mem_wready_i   => axi_wready_w,
               mem_bvalid_i   => axi_bvalid_w,
               --
               mem_awvalid_o  => axi_awvalid_w,
               mem_awaddr_o   => axi_awaddr_w,
               mem_wvalid_o   => axi_wvalid_w,
               mem_wdata_o    => axi_wdata_w,
               mem_wstrb_o    => axi_wstrb_w,
               mem_bready_o   => axi_bready_w,
               --
               mem_arready_i  => axi_arready_w,
               mem_rvalid_i   => axi_rvalid_w,
               mem_rdata_i    => axi_rdata_w,
               --
               mem_arvalid_o  => axi_arvalid_w,
               mem_araddr_o   => axi_araddr_w,
               mem_rready_o   => axi_rready_w,
               --  pp io
               uart_rxd_i     => uart_txd_i,
               uart_txd_o     => dbg_txd_w,
               rv_core_ena_o  => rv_core_ena_w );         -- cpu / cache enable


  xrst_cpu_w   <= rv_core_ena_w and xrst_i;


  p_txd : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        txd_q <= '1';
      else
        txd_q <= dbg_txd_w and uart_txd_w;
      end if;
    end if;
  end process p_txd;

  uart_rxd_o     <= txd_q;

end rtl;
