-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


-- - - - - - - - - - - - - - - - - - - - -
-- module:  general purpose IO peripheral
-- - - - - - - - - - - - - - - - - - - - -
entity vsgpio is
  generic ( NR_GPIOS  : integer range 1 to 32 := 8 );  
  port ( clk_i        : in  std_logic;
         xrst_i       : in  std_logic;
         --  vsbus,  inpts retimed in higher level
         pp_rd_i      : in  std_logic;
         pp_wr_i      : in  slv_03_00;   --! 3:1 NC
         pp_addr_i    : in  slv_31_00;
         pp_data_i    : in  slv_31_00;
         pp_data_o    : out slv_31_00;
         --  interrupt request
         intr_o       : out std_logic;
         --  gpios to pads
         gpio_i       : in  slv_t(NR_GPIOS-1 downto 0);
         gpio_o       : out slv_t(NR_GPIOS-1 downto 0);
         gpio_oena_o  : out slv_t(NR_GPIOS-1 downto 0)
         );
end entity vsgpio;


architecture rtl of vsgpio is

  subtype gp_slv_t is slv_t(NR_GPIOS-1 downto 0);

  -- - - - - - - - - - - - - - - - - - - - -
  -- gpio defs    - default, range
  -- - - - - - - - - - - - - - - - - - - - -
  constant GPIO_DIRECTION                      : slv_07_00 := X"00";
  constant GPIO_DIRECTION_OUTPUT_DEFAULT       : std_logic  := '0';
  constant GPIO_DIRECTION_OUTPUT_R             : gp_slv_t := (others => '0');

  constant GPIO_INPUT                          : slv_07_00 := X"04";
  constant GPIO_INPUT_VALUE_R                  : gp_slv_t := (others => '0');

  constant GPIO_OUTPUT                         : slv_07_00 := X"08";
  constant GPIO_OUTPUT_DATA_R                  : gp_slv_t := (others => '0');

  constant GPIO_OUTPUT_SET                     : slv_07_00 := X"0C";
  constant GPIO_OUTPUT_SET_DATA_R              : gp_slv_t := (others => '0');

  constant GPIO_OUTPUT_CLR                     : slv_07_00 := X"10";
  constant GPIO_OUTPUT_CLR_DATA_R              : gp_slv_t := (others => '0');

  constant GPIO_INT_MASK                       : slv_07_00 := X"14";
  constant GPIO_INT_MASK_ENABLE_DEFAULT        : std_logic  := '0';
  constant GPIO_INT_MASK_ENABLE_R              : gp_slv_t := (others => '0');

  constant GPIO_INT_SET                        : slv_07_00 := X"18";
  constant GPIO_INT_SET_SW_IRQ_R               : gp_slv_t := (others => '0');

  constant GPIO_INT_CLR                        : slv_07_00 := X"1C";
  constant GPIO_INT_CLR_ACK_R                  : gp_slv_t := (others => '0');

  constant GPIO_INT_STATUS                     : slv_07_00 := X"20";
  constant GPIO_INT_STATUS_RAW_R               : gp_slv_t := (others => '0');

  constant GPIO_INT_LEVEL                      : slv_07_00 := X"24";
  constant GPIO_INT_LEVEL_ACTIVE_HIGH_DEFAULT  : std_logic  := '0';
  constant GPIO_INT_LEVEL_ACTIVE_HIGH_R        : gp_slv_t := (others => '0');

  constant GPIO_INT_MODE                       : slv_07_00 := X"28";
  constant GPIO_INT_MODE_EDGE_DEFAULT          : std_logic  := '0';
  constant GPIO_INT_MODE_EDGE_R                : gp_slv_t := (others => '0');


  --  gpio_direction_output (internal)
  signal gpio_direction_output_q           : gp_slv_t;

  signal gpio_output_set_wr_w              : std_logic;
  signal gpio_output_clr_wr_w              : std_logic;
  signal gpio_output_wr_w                  : std_logic;
  signal gpio_output_set_data_out_w        : gp_slv_t;
  signal gpio_output_clr_data_out_w        : gp_slv_t;
  signal gpio_output_data_out_w            : gp_slv_t;

  signal gpio_int_mask_enable_q            : gp_slv_t;
  signal gpio_int_mask_enable_out_w        : gp_slv_t;

  signal gpio_int_set_wr_w                 : std_logic;
  signal gpio_int_clr_wr_w                 : std_logic;
  signal gpio_int_set_sw_irq_out_w         : gp_slv_t;
  signal gpio_int_clr_ack_out_w            : gp_slv_t;
  --  gpio_int_level_active_high (internal)
  signal gpio_int_level_active_high_q      : gp_slv_t;
  signal gpio_int_level_active_high_out_w  : gp_slv_t;
  --  gpio_int_mode_edge (internal)
  signal gpio_int_mode_edge_q              : gp_slv_t;

  signal gpio_int_mode_edge_out_w          : gp_slv_t;
  signal gpio_input_value_in_w             : gp_slv_t;
  signal gpio_output_data_in_w             : gp_slv_t;
  signal gpio_int_status_raw_in_w          : gp_slv_t;

  signal data_w                            : slv_t(pp_data_o'range);

  signal input_ms                          : gp_slv_t;
  signal input_q                           : gp_slv_t;

  signal output_q                          : gp_slv_t;
  signal output_next_w                     : gp_slv_t;

  signal intr_q                            : std_logic;

  signal interrupt_raw_q                   : gp_slv_t;
  signal interrupt_raw_w                   : gp_slv_t;

  signal input_last_q                      : gp_slv_t;

  signal active_low_w                      : gp_slv_t;
  signal falling_edge_w                    : gp_slv_t;
  signal level_mask_w                      : gp_slv_t;
  signal edge_mask_w                       : gp_slv_t;

  signal level_active_w                    : gp_slv_t;

  signal edge_detect_w                     : gp_slv_t;
  signal edge_active_w                     : gp_slv_t;

begin

  -- - - - - - - - - - - - - - - - - -
  --  request logic
  -- - - - - - - - - - - - - - - - - -

  --  gpio_direction_output (internal)
  p_gp_wr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        gpio_direction_output_q <= (others => GPIO_DIRECTION_OUTPUT_DEFAULT);
        gpio_int_mask_enable_q  <= (others => GPIO_INT_MASK_ENABLE_DEFAULT);
         gpio_int_level_active_high_q <=
          (others => GPIO_INT_LEVEL_ACTIVE_HIGH_DEFAULT);
        gpio_int_mode_edge_q    <= (others => GPIO_INT_MODE_EDGE_DEFAULT);
     else
        --
         --
        if (pp_wr_i(0) = '1') then
          case pp_addr_i(7 downto 0) is
            when GPIO_DIRECTION =>
              gpio_direction_output_q <=
                pp_data_i(GPIO_DIRECTION_OUTPUT_R'range);
            when GPIO_INT_MASK =>
              gpio_int_mask_enable_q <= pp_data_i(GPIO_INT_MASK_ENABLE_R'range);
            when GPIO_INT_LEVEL =>
              gpio_int_level_active_high_q <=
                pp_data_i(GPIO_INT_LEVEL_ACTIVE_HIGH_R'range);
            when GPIO_INT_MODE =>            -- gpio_int_mode_edge (internal)
              gpio_int_mode_edge_q   <= pp_data_i(GPIO_INT_MODE_EDGE_R'range);
            when others =>
              null;
          end case;
        end if;
      end if;
    end if;
  end process p_gp_wr;

  --  register gpio_output
  gpio_output_wr_w     <= pp_wr_i(0) and eq(pp_addr_i(7 downto 0), GPIO_OUTPUT);
  --  register gpio_output_clr
  gpio_output_clr_wr_w <= pp_wr_i(0) and
                          eq(pp_addr_i(7 downto 0), GPIO_OUTPUT_CLR);

  --  register gpio_output_set
  gpio_output_set_wr_w <= pp_wr_i(0) and
                          eq(pp_addr_i(7 downto 0), GPIO_OUTPUT_SET);

  -- register gpio_int_set
  gpio_int_set_wr_w    <= pp_wr_i(0) and eq(pp_addr_i(7 downto 0),GPIO_INT_SET);

  --  register gpio_int_clr
  gpio_int_clr_wr_w    <= pp_wr_i(0) and
                          eq(pp_addr_i(7 downto 0), GPIO_INT_CLR);
  

  --  gpio_int_mask_enable (internal)
  gpio_int_mask_enable_out_w       <= gpio_int_mask_enable_q;

  gpio_int_level_active_high_out_w <= gpio_int_level_active_high_q;

  --  gpio_int_mode_edge (internal)
  gpio_int_mode_edge_out_w         <= gpio_int_mode_edge_q;

  gpio_output_data_in_w            <= output_q;


  -- - - - - - - - - - - - - - - - - -
  --  read mux
  -- - - - - - - - - - - - - - - - - -
  p_c_omux : process (pp_rd_i, pp_addr_i, gpio_direction_output_q,
                      gpio_input_value_in_w, gpio_output_data_in_w,
                      gpio_int_mask_enable_q, gpio_int_status_raw_in_w,
                      gpio_int_level_active_high_q, gpio_int_mode_edge_q)
  begin
    --
    data_w <= (others => '0');
    --
    if (pp_rd_i = '1') then
      case pp_addr_i(7 downto 0) is
        when GPIO_DIRECTION =>
          data_w(GPIO_DIRECTION_OUTPUT_R'range)     <= gpio_direction_output_q;
        when GPIO_INPUT =>
          data_w(GPIO_INPUT_VALUE_R'range)          <= gpio_input_value_in_w;
        when GPIO_OUTPUT =>
          data_w(GPIO_OUTPUT_DATA_R'range)          <= gpio_output_data_in_w;
        when GPIO_INT_MASK =>
          data_w(GPIO_INT_MASK_ENABLE_R'range)      <= gpio_int_mask_enable_q;
        when GPIO_INT_STATUS =>
          data_w(GPIO_INT_STATUS_RAW_R'range)       <= gpio_int_status_raw_in_w;
        when GPIO_INT_LEVEL =>
          data_w(GPIO_INT_LEVEL_ACTIVE_HIGH_R'range) <=
            gpio_int_level_active_high_q;
        when GPIO_INT_MODE =>
          data_w(GPIO_INT_MODE_EDGE_R'range)        <= gpio_int_mode_edge_q;
        when others  =>
          null;
      end case;
    end if;
  end process p_c_omux;


  pp_data_o <= data_w;



  -- - - - - - - - - - - - - - - - - -
  --  inputs resynced
  -- - - - - - - - - - - - - - - - - -
  --  resync inputs
  p_gpin : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        input_ms <= (NR_GPIOS-1 downto 0 => '0');
        input_q  <= (NR_GPIOS-1 downto 0 => '0');
      else
        input_ms <= gpio_i;
        input_q  <= input_ms;
      end if;
    end if;
  end process p_gpin;

  gpio_input_value_in_w <= input_q;


  -- - - - - - - - - - - - - - - - - -
  --  outputs
  -- - - - - - - - - - - - - - - - - -
  gpio_output_set_data_out_w <= pp_data_i(GPIO_OUTPUT_SET_DATA_R'range);
  gpio_output_clr_data_out_w <= pp_data_i(GPIO_OUTPUT_CLR_DATA_R'range);
  gpio_output_data_out_w     <= pp_data_i(GPIO_OUTPUT_DATA_R'range);

  p_c_gponxt : process ( output_q,
                         gpio_output_set_wr_w, gpio_output_set_data_out_w,
                         gpio_output_clr_wr_w, gpio_output_clr_data_out_w,
                         gpio_output_wr_w, gpio_output_data_out_w)
  begin
    output_next_w <= output_q;
    if (gpio_output_set_wr_w = '1') then    -- gpio_output_set_data (external)
        output_next_w  <= output_q or gpio_output_set_data_out_w;
    elsif (gpio_output_clr_wr_w = '1') then -- gpio_output_clr_data (external)
      output_next_w    <= output_q and (not gpio_output_clr_data_out_w);
    elsif (gpio_output_wr_w = '1') then     -- gpio_output_data (external)
        output_next_w  <= gpio_output_data_out_w;
    end if;
  end process p_c_gponxt;


  p_gpout : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        output_q  <= (NR_GPIOS-1 downto 0 => '0');
      else
        output_q  <= output_next_w;
      end if;
    end if;
  end process p_gpout;

  gpio_o <= output_q;


  -- - - - - - - - - - - - - - - - - -
  --  interrupts
  -- - - - - - - - - - - - - - - - - -


  p_ilast : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        input_last_q  <= (NR_GPIOS-1 downto 0 => '0');
      else
        input_last_q  <= gpio_input_value_in_w;
      end if;
    end if;
  end process p_ilast;


  active_low_w   <= (not gpio_int_level_active_high_out_w);
  falling_edge_w <= (not gpio_int_level_active_high_out_w);
  level_mask_w   <= (not gpio_int_mode_edge_out_w);
  edge_mask_w    <= gpio_int_mode_edge_out_w;

  level_active_w <= (gpio_input_value_in_w xor active_low_w) and level_mask_w;

  edge_detect_w  <= (input_last_q xor gpio_input_value_in_w);
  edge_active_w  <= (edge_detect_w and
                     (gpio_input_value_in_w xor falling_edge_w)) and
                    edge_mask_w;


  -- gpio_int_set_sw_irq (external)
  gpio_int_set_sw_irq_out_w  <= pp_data_i(GPIO_INT_SET_SW_IRQ_R'range);
 
  -- gpio_int_clr_ack (external)
  gpio_int_clr_ack_out_w     <= pp_data_i(GPIO_INT_CLR_ACK_R'range);



  p_c_intr : process (interrupt_raw_q, gpio_int_clr_wr_w,
                      gpio_int_clr_ack_out_w, level_active_w, edge_active_w,
                      gpio_int_set_wr_w, gpio_int_set_sw_irq_out_w)
    variable interrupt_raw_v : slv_t(interrupt_raw_w'range);
  begin
    interrupt_raw_v := interrupt_raw_q;
      --  clear (ACK)
    if (gpio_int_clr_wr_w = '1') then
      interrupt_raw_v := interrupt_raw_v and (not gpio_int_clr_ack_out_w);
    end if;
    --  new interrupts
    interrupt_raw_v := interrupt_raw_v or level_active_w or edge_active_w;
    --  set (SW IRQ)
    if (gpio_int_set_wr_w = '1') then -- gpio_int_set_sw_irq (external)
      interrupt_raw_v := interrupt_raw_v or gpio_int_set_sw_irq_out_w;
    end if;
    interrupt_raw_w <= interrupt_raw_v;
  end process p_c_intr;



  p_intr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        intr_q          <= '0';
        interrupt_raw_q <= (NR_GPIOS-1 downto 0 => '0');
      else
        if (unsigned(interrupt_raw_q and gpio_int_mask_enable_out_w) /= 0) then
          intr_q           <= '1';
        else
          intr_q           <= '0';
        end if;
        interrupt_raw_q  <= interrupt_raw_w;
      end if;
    end if;
  end process p_intr;

  gpio_int_status_raw_in_w <= interrupt_raw_q;
  intr_o                   <= intr_q;


  gpio_oena_o <= gpio_direction_output_q;

end rtl;
