-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


-- - - - - - - - - - - - - - - - - - - - -
-- module:  system tick timer
-- - - - - - - - - - - - - - - - - - - - -
entity vstimer is
  port ( clk_i      : in  std_logic;
         xrst_i     : in  std_logic;
         --  vsbus
         pp_addr_i  : in  slv_31_00;
         pp_data_i  : in  slv_31_00;
         pp_rd_i    : in  std_logic;
         pp_wr_i    : in  slv_03_00;   --! 3:1 NC
         pp_data_o  : out slv_31_00;
         intr_o     : out std_logic
         );
end entity vstimer;


architecture rtl of vstimer is

  -- - - - - - - - - - - - - - - - - - - - -
  --  timer defs   - default, bit, range
  -- - - - - - - - - - - - - - - - - - - - -
  constant TIMER_CTRL0                    : slv_08_00 := "000001000";   -- 0x008
  constant TIMER_CTRL0_INTERRUPT_DEFAULT  : std_logic := '0';
  constant TIMER_CTRL0_INTERRUPT_B        : natural   := 1;
  constant TIMER_CTRL0_ENABLE_DEFAULT     : std_logic := '0';
  constant TIMER_CTRL0_ENABLE_B           : natural   := 2;

  constant TIMER_CMP0                     : slv_08_00 := "000001100";   -- 0x00C
  constant TIMER_CMP0_VALUE_DEFAULT       : std_logic := '0';
  constant TIMER_CMP0_VALUE_R             : slv_31_00 := (others => '0');

  constant TIMER_VAL0 : slv_08_00         := "000010000"; -- 0x010
  constant TIMER_VAL0_CURRENT_R           : slv_31_00 := (others => '0');

  constant TIMER_CTRL1                    : slv_08_00 := "000010100";   -- 0x014
  constant TIMER_CTRL1_INTERRUPT_DEFAULT  : std_logic := '0';
  constant TIMER_CTRL1_INTERRUPT_B        : natural   := 1;

  constant TIMER_CTRL1_ENABLE             : natural   := 2;
  constant TIMER_CTRL1_ENABLE_DEFAULT     : std_logic := '0';
  constant TIMER_CTRL1_ENABLE_B           : natural   := 2;

  constant TIMER_CMP1                     : slv_08_00 := "000011000";   -- 0x018
  constant TIMER_CMP1_VALUE_DEFAULT       : std_logic := '0';
  constant TIMER_CMP1_VALUE_R             : slv_31_00 := (others => '0');

  constant TIMER_VAL1                     : slv_08_00 := "000011100";   -- 0x01C
  constant TIMER_VAL1_CURRENT_R           : slv_31_00 := (others => '0');


  -- timer_ctrl0_interrupt (internal)
  signal timer_ctrl0_interrupt_q      : std_logic;
  signal timer_ctrl0_interrupt_out_w  : std_logic;
    -- timer_ctrl0_enable (internal)
  signal timer_ctrl0_enable_q         : std_logic;
  signal timer_ctrl0_enable_out_w     : std_logic;

  signal timer_cmp0_value_q           : slv_31_00;
  signal timer_cmp0_value_out_w       : slv_31_00;

  signal timer_val0_wr_q              : std_logic;
  signal timer_val0_current_out_w     : slv_31_00;
  -- timer_ctrl1_interrupt (internal)
  signal timer_ctrl1_interrupt_q      : std_logic;

  signal timer_ctrl1_interrupt_out_w  : std_logic;
  -- timer_ctrl1_enable (internal)
  signal timer_ctrl1_enable_q         : std_logic;
  signal timer_ctrl1_enable_out_w     : std_logic;

  -- timer_cmp1_value (internal)
  signal timer_cmp1_value_q           : slv_31_00;
  signal timer_cmp1_value_out_w       : slv_31_00;

  signal timer_val1_current_out_w     : slv_31_00;
  signal timer_val0_current_in_w      : slv_31_00;
  signal timer_val1_current_in_w      : slv_31_00;
  signal data_w                       : slv_31_00;
  signal timer_val0_wr_req_w          : std_logic;
  signal timer0_value_q               : slv_31_00;
  signal timer0_irq_w                 : std_logic;
  signal timer1_value_q               : slv_31_00;
  signal timer1_irq_w                 : std_logic;
  signal intr_q                       : std_logic;

begin


  --  timer_ctrl0_interrupt (internal)
  p_tctrl0_i : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_ctrl0_interrupt_q <= TIMER_CTRL0_INTERRUPT_DEFAULT;
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CTRL0)) then
          timer_ctrl0_interrupt_q <= pp_data_i(TIMER_CTRL0_INTERRUPT_B);
        end if;
      end if;
    end if;
  end process p_tctrl0_i;


  timer_ctrl0_interrupt_out_w <= timer_ctrl0_interrupt_q;


  --  timer_ctrl0_enable (internal)
  p_tctrl0_en : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_ctrl0_enable_q <= TIMER_CTRL0_ENABLE_DEFAULT;
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CTRL0)) then
          timer_ctrl0_enable_q <= pp_data_i(TIMER_CTRL0_ENABLE_B);
        end if;
      end if;
    end if;
  end process p_tctrl0_en;


  timer_ctrl0_enable_out_w <= timer_ctrl0_enable_q;


  --  timer_cmp0_value (internal)
  p_tcmp0_val : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_cmp0_value_q <= (others => TIMER_CMP0_VALUE_DEFAULT);
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CMP0)) then
          timer_cmp0_value_q <= pp_data_i(TIMER_CMP0_VALUE_R'range);
        end if;
      end if;
    end if;
  end process p_tcmp0_val;


  timer_cmp0_value_out_w <= timer_cmp0_value_q;


  --  timer_ctrl1_interrupt (internal)
  p_tctrl1_iq : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_ctrl1_interrupt_q <= TIMER_CTRL1_INTERRUPT_DEFAULT;
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CTRL1)) then
          timer_ctrl1_interrupt_q <= pp_data_i(TIMER_CTRL1_INTERRUPT_B);
        end if;
      end if;
    end if;
  end process p_tctrl1_iq;


  timer_ctrl1_interrupt_out_w <= timer_ctrl1_interrupt_q;


  --  timer_ctrl1_enable (internal)
  p_tctrl1_en : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_ctrl1_enable_q <= TIMER_CTRL1_ENABLE_DEFAULT;
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CTRL1)) then
          timer_ctrl1_enable_q <= pp_data_i(TIMER_CTRL1_ENABLE_B);
        end if;
      end if;
    end if;
  end process p_tctrl1_en;


  timer_ctrl1_enable_out_w <= timer_ctrl1_enable_q;


  --  timer_cmp1_value (internal)
  p_tcmp1_val : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer_cmp1_value_q <= (others => TIMER_CMP1_VALUE_DEFAULT);
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_CMP1)) then
          timer_cmp1_value_q <= pp_data_i(TIMER_CMP1_VALUE_R'range);
        end if;
      end if;
    end if;
  end process p_tcmp1_val;


  timer_cmp1_value_out_w <= timer_cmp1_value_q;


  -- - - - - - - - - - - - - - - - - - - - -
  --  read mux
  -- - - - - - - - - - - - - - - - - - - - -
  p_c_omux : process (pp_addr_i, timer_ctrl0_interrupt_q,
                      timer_ctrl0_enable_q, timer_cmp0_value_q,
                      timer_val0_current_in_w, timer_ctrl1_interrupt_q,
                      timer_ctrl1_enable_q, timer_cmp1_value_q,
                      timer_val1_current_in_w)
  begin
    --
    data_w <= X"00000000";
    --
    case pp_addr_i(8 downto 0) is
      when TIMER_CTRL0 =>
        data_w(TIMER_CTRL0_INTERRUPT_B)    <= timer_ctrl0_interrupt_q;
        data_w(TIMER_CTRL0_ENABLE_B)       <= timer_ctrl0_enable_q;
      when TIMER_CMP0 =>
        data_w(TIMER_CMP0_VALUE_R'range)   <= timer_cmp0_value_q;
      when TIMER_VAL0 =>
        data_w(TIMER_VAL0_CURRENT_R'range) <= timer_val0_current_in_w;
      when TIMER_CTRL1 =>
        data_w(TIMER_CTRL1_INTERRUPT_B)    <= timer_ctrl1_interrupt_q;
        data_w(TIMER_CTRL1_ENABLE_B)       <= timer_ctrl1_enable_q;
      when TIMER_CMP1 =>
        data_w(TIMER_CMP1_VALUE_R'range)   <= timer_cmp1_value_q;
      when TIMER_VAL1 =>
        data_w(TIMER_VAL1_CURRENT_R'range) <= timer_val1_current_in_w;
      when others =>
        data_w <= X"00000000";
    end case;
  end process p_c_omux;


  pp_data_o  <= data_w when (pp_rd_i = '1') else X"00000000";


  -- - - - - - - - - - - - - - - - - - - - -
  --  timer0
  -- - - - - - - - - - - - - - - - - - - - -
  p_tim0 : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer0_value_q <= X"00000000";
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_VAL0)) then
          timer0_value_q <= timer_val0_current_out_w;
        elsif (timer_ctrl0_enable_out_w = '1') then
          timer0_value_q <= inc1u(timer0_value_q);
        end if;
      end if;
    end if;
  end process p_tim0;


  timer_val0_current_in_w <= timer0_value_q;

  timer0_irq_w <= '1' when ((timer_val0_current_in_w = timer_cmp0_value_out_w)
                            and timer_ctrl0_interrupt_out_w = '1' and
                            timer_ctrl0_enable_out_w = '1')
                  else '0';

  -- timer_val0_current (external)
  timer_val0_current_out_w <= pp_data_i(TIMER_VAL0_CURRENT_R'range);


  -- - - - - - - - - - - - - - - - - - - - -
  --  timer1
  -- - - - - - - - - - - - - - - - - - - - -
  p_tim1 : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        timer1_value_q <= X"00000000";
      else
        if (pp_wr_i(0) = '1' and (pp_addr_i(8 downto 0) = TIMER_VAL1)) then
          timer1_value_q <= timer_val1_current_out_w;
        elsif (timer_ctrl1_enable_out_w = '1') then
          timer1_value_q <= inc1u(timer1_value_q);
        end if;
      end if;
    end if;
  end process p_tim1;


  timer_val1_current_in_w <= timer1_value_q;

  timer1_irq_w <= '1' when ((timer_val1_current_in_w = timer_cmp1_value_out_w)
                            and timer_ctrl1_interrupt_out_w = '1' and
                            timer_ctrl1_enable_out_w = '1')
                  else '0';


  -- timer_val1_current (external)
  timer_val1_current_out_w <= pp_data_i(TIMER_VAL1_CURRENT_R'range);


  -- - - - - - - - - - - - - - - - - - - - -
  --  IRQ output
  -- - - - - - - - - - - - - - - - - - - - -
  p_intr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        intr_q <= '0';
      else
        if (timer0_irq_w = '1' or timer1_irq_w = '1') then
          intr_q <= '1';
        else
          intr_q <= '0';
        end if;
      end if;
    end if;
  end process p_intr;


  intr_o <= intr_q;


 end rtl;
