-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity dcache is
  generic (AXI_ID      : slv_03_00 := X"0";
           WAYS        : natural range 1 to 8  := 2;   -- Number of ways 1,2,4,8
           LINE_ADDR_W : natural range 6 to 12 := 8;   -- Line address
           LINE_SIZE_W : natural range 2 to 8  := 5 ); -- Line size (byte addr)
           --
           -- TAG_ADDR_W:  32 -LINE_ADDR_W -LINE_SIZE_W = 32 - 8 - 5 = 19
           --
  port ( clk_i             : in  std_logic;
         xrst_i            : in  std_logic;       -- syncr reset
         --
         cpu_rd_i          : in  std_logic;       --  mem inputs (from core)
         cpu_wr_i          : in  slv_03_00;
         cpu_addr_i        : in  slv_31_00;       -- BYTE addr
         cpu_data_wr_i     : in  slv_31_00;
         cpu_req_tag_i     : in  slv_10_00;
         cpu_invalidate_i  : in  std_logic;
         cpu_writeback_i   : in  std_logic;
         cpu_flush_i       : in  std_logic;
         cpu_data_rd_o     : out slv_31_00;       --  mem outputs (to core)
         cpu_accept_o      : out std_logic;
         cpu_ack_o         : out std_logic;
         cpu_error_o       : out std_logic;
         cpu_resp_tag_o    : out slv_10_00;
         --
         axi_awready_i     : in  std_logic;       -- axi if inputs (from DDR)
         axi_wready_i      : in  std_logic;
         axi_bvalid_i      : in  std_logic;
         axi_bresp_i       : in  slv_01_00;
         axi_bid_i         : in  slv_03_00;
         axi_arready_i     : in  std_logic;
         axi_rvalid_i      : in  std_logic;
         axi_rdata_i       : in  slv_31_00;
         axi_rresp_i       : in  slv_01_00;
         axi_rid_i         : in  slv_03_00;
         axi_rlast_i       : in  std_logic;
         axi_awvalid_o     : out std_logic;       -- axi if outputs (to DDR)
         axi_awaddr_o      : out slv_31_00;
         axi_awid_o        : out slv_03_00;
         axi_awlen_o       : out slv_07_00;
         axi_awburst_o     : out slv_01_00;
         axi_wvalid_o      : out std_logic;
         axi_wdata_o       : out slv_31_00;
         axi_wstrb_o       : out slv_03_00;
         axi_wlast_o       : out std_logic;
         axi_bready_o      : out std_logic;
         axi_arvalid_o     : out std_logic;
         axi_araddr_o      : out slv_31_00;
         axi_arid_o        : out slv_03_00;
         axi_arlen_o       : out slv_07_00;
         axi_arburst_o     : out slv_01_00;
         axi_rready_o      : out std_logic
         );
end entity dcache;


architecture rtl of dcache is

  component dcache_core
    generic (WAYS        : natural range 1 to 8  := 2;   -- Nro of ways 1,2,4,8
             LINE_ADDR_W : natural range 6 to 12 := 8;   -- Line address
             LINE_SIZE_W : natural range 2 to 8  := 5 ); -- Line size, byte addr
    port ( clk_i                : in  std_logic;
           xrst_i               : in  std_logic;         -- syncr reset
           -- CPU
           cpu_rd_i          : in  std_logic;
           cpu_wr_i          : in  slv_03_00;
           cpu_addr_i        : in  slv_31_00;            -- WORD address
           cpu_wdata_i       : in  slv_31_00;
           cpu_req_tag_i     : in  slv_10_00;
           cpu_invalidate_i  : in  std_logic;
           cpu_wb_i          : in  std_logic;
           cpu_flush_i       : in  std_logic;
           cpu_rdata_o       : out slv_31_00;
           cpu_accept_o      : out std_logic;
           cpu_ack_o         : out std_logic;
           cpu_error_o       : out std_logic;
           cpu_resp_tag_o    : out slv_10_00;
           -- XMEM
           xmem_accept_i     : in  std_logic;
           xmem_ack_i        : in  std_logic;
           xmem_error_i      : in  std_logic;
           xmem_rdata_i      : in  slv_31_00;
           xmem_rd_o         : out std_logic;
           xmem_wr_o         : out slv_03_00;
           xmem_len_o        : out slv_07_00;
           xmem_addr_o       : out slv_31_00;
           xmem_wdata_o      : out slv_31_00 );
  end component;


  component dcache_axi
    generic ( AXI_ID    : slv_03_00 := X"0" );
    port ( clk_i          : in     std_logic;
           xrst_i         : in     std_logic;
           --
           mem_rd_i       : in     std_logic;
           mem_wr_i       : in     std_logic;
           mem_len_i      : in     slv_07_00;
           mem_addr_i     : in     slv_31_00;
           mem_wdata_i    : in     slv_31_00;
           mem_accept_o   : buffer std_logic;
           mem_ack_o      : buffer std_logic;
           mem_rdata_o    : out    slv_31_00;
           mem_error_o    : out    std_logic;
           -- read req
           axi_arvalid_o  : out    std_logic;
           axi_araddr_o   : out    slv_31_00;
           axi_arid_o     : out    slv_03_00;
           axi_arlen_o    : out    slv_07_00;
           axi_arburst_o  : out    slv_01_00;
           axi_rready_o   : out    std_logic;
           -- read resp
           axi_arready_i  : in     std_logic;
           axi_rvalid_i   : in     std_logic;
           axi_rdata_i    : in     slv_31_00;
           axi_rresp_i    : in     slv_01_00;
           axi_rid_i      : in     slv_03_00;
           axi_rlast_i    : in     std_logic;
           -- write req
           axi_awready_i  : in     std_logic;
           axi_wready_i   : in     std_logic;
           axi_bvalid_i   : in     std_logic;
           axi_bresp_i    : in     slv_01_00;
           axi_bid_i      : in     slv_03_00;
           --
           axi_awvalid_o  : buffer std_logic;
           axi_awaddr_o   : out    slv_31_00;
           axi_awid_o     : out    slv_03_00;
           axi_awlen_o    : out    slv_07_00;
           axi_awburst_o  : out    slv_01_00;
           axi_wvalid_o   : buffer std_logic;
           axi_wdata_o    : out    slv_31_00;
           axi_wstrb_o    : out    slv_31_00;
           axi_wlast_o    : buffer std_logic;
           axi_bready_o   : buffer std_logic );
  end component;


  signal xmem_len_w       : slv_07_00;
  signal xmem_error_w     : std_logic;
  signal xmem_addr_w      : slv_31_00;
  signal xmem_ack_w       : std_logic;
  signal xmem_wdata_w     : slv_31_00;
  signal xmem_rd_w        : std_logic;
  signal xmem_wr_w        : slv_03_00;
  signal xmem_accept_w    : std_logic;
  signal xmem_rdata_w     : slv_31_00;

  signal axi_awvalid_o_w  : std_logic;
  signal axi_wvalid_o_w   : std_logic;
  signal axi_bready_o_w   : std_logic;

  signal axi_wstrb_o_w    : slv_31_00;
  signal axi_wlast_w      : std_logic;


begin

  u_core : dcache_core
    generic  map ( WAYS        => WAYS,           -- Number of ways 1,2,4,8
                   LINE_ADDR_W => LINE_ADDR_W,    -- Line address
                   LINE_SIZE_W => LINE_SIZE_W )   -- Line size, byte addr
    port map ( clk_i             => clk_i,
               xrst_i            => xrst_i,
               --
               cpu_rd_i          => cpu_rd_i,
               cpu_wr_i          => cpu_wr_i,
               cpu_addr_i        => cpu_addr_i,           -- BYTE addr, [1:0] NC
               cpu_wdata_i       => cpu_data_wr_i,
               cpu_req_tag_i     => cpu_req_tag_i,
               cpu_invalidate_i  => cpu_invalidate_i,
               cpu_wb_i          => cpu_writeback_i,
               cpu_flush_i       => cpu_flush_i,
               cpu_rdata_o       => cpu_data_rd_o,
               cpu_accept_o      => cpu_accept_o,
               cpu_ack_o         => cpu_ack_o,
               cpu_error_o       => cpu_error_o,
               cpu_resp_tag_o    => cpu_resp_tag_o,       -- mem resp tag
               --
               xmem_accept_i     => xmem_accept_w,
               xmem_ack_i        => xmem_ack_w,
               xmem_error_i      => xmem_error_w,
               xmem_rdata_i      => xmem_rdata_w,
               xmem_rd_o         => xmem_rd_w,
               xmem_wr_o         => xmem_wr_w,
               xmem_len_o        => xmem_len_w,
               xmem_addr_o       => xmem_addr_w,
               xmem_wdata_o      => xmem_wdata_w );


  u_axi : dcache_axi
    generic map (AXI_ID => AXI_ID )
    port map ( clk_i          => clk_i,
               xrst_i         => xrst_i,
               --
               mem_rd_i       => xmem_rd_w,
               mem_wr_i       => xmem_wr_w(0),            -- NC remove 3:1
               mem_len_i      => xmem_len_w,
               mem_addr_i     => xmem_addr_w,             -- BYTE, [1:0] NC
               mem_wdata_i    => xmem_wdata_w,
               mem_accept_o   => xmem_accept_w,
               mem_ack_o      => xmem_ack_w,
               mem_rdata_o    => xmem_rdata_w,
               mem_error_o    => xmem_error_w,
               -- read req
               axi_arvalid_o  => axi_arvalid_o,
               axi_araddr_o   => axi_araddr_o,
               axi_arid_o     => axi_arid_o,
               axi_arlen_o    => axi_arlen_o,
               axi_arburst_o  => axi_arburst_o,
               axi_rready_o   => axi_rready_o,
               -- read resp
               axi_arready_i  => axi_arready_i,
               axi_rvalid_i   => axi_rvalid_i,
               axi_rdata_i    => axi_rdata_i,
               axi_rresp_i    => axi_rresp_i,
               axi_rid_i      => axi_rid_i,
               axi_rlast_i    => axi_rlast_i,
               -- write req
               axi_awready_i  => axi_awready_i,
               axi_wready_i   => axi_wready_i,
               axi_bvalid_i   => axi_bvalid_i,
               axi_bresp_i    => axi_bresp_i,
               axi_bid_i      => axi_bid_i,
               -- write resp
               axi_awvalid_o  => axi_awvalid_o_w,
               axi_awaddr_o   => axi_awaddr_o,
               axi_awid_o     => axi_awid_o,
               axi_awlen_o    => axi_awlen_o,
               axi_awburst_o  => axi_awburst_o,
               axi_wvalid_o   => axi_wvalid_o_w,
               axi_wdata_o    => axi_wdata_o,
               axi_wstrb_o    => axi_wstrb_o_w,           -- NC 31:4
               axi_wlast_o    => axi_wlast_w,
               axi_bready_o   => axi_bready_o_w );


  axi_awvalid_o  <= axi_awvalid_o_w;
  axi_wvalid_o   <= axi_wvalid_o_w;
  axi_bready_o   <= axi_bready_o_w;
  axi_wstrb_o    <= axi_wstrb_o_w(3 downto 0);
  axi_wlast_o    <= axi_wlast_w;

end rtl;
