-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity dcache_axi is
  generic ( AXI_ID  : slv_03_00 := X"0" );
  port ( clk_i          : in     std_logic;
         xrst_i         : in     std_logic;
         --
         mem_accept_o   : buffer std_logic;
         mem_ack_o      : buffer std_logic;
         mem_rdata_o    : out    slv_31_00;
         mem_wr_i       : in     std_logic;    -- high during whole burst time
         mem_rd_i       : in     std_logic;
         mem_addr_i     : in     slv_31_00;    -- [1:0] NC
         mem_wdata_i    : in     slv_31_00;    -- 32b
         mem_len_i      : in     slv_07_00;    -- 8b
         mem_error_o    : out    std_logic;
         -- read req
         axi_arvalid_o  : out    std_logic;
         axi_araddr_o   : out    slv_31_00;
         axi_arid_o     : out    slv_03_00;
         axi_arlen_o    : out    slv_07_00;
         axi_arburst_o  : out    slv_01_00;
         axi_rready_o   : out    std_logic;
         -- read resp
         axi_arready_i  : in     std_logic;
         axi_rvalid_i   : in     std_logic;
         axi_rdata_i    : in     slv_31_00;    -- 32b
         axi_rresp_i    : in     slv_01_00;
         axi_rid_i      : in     slv_03_00;
         axi_rlast_i    : in     std_logic;
         -- write req
         axi_awready_i  : in     std_logic;
         axi_wready_i   : in     std_logic;
         axi_bvalid_i   : in     std_logic;
         axi_bresp_i    : in     slv_01_00;
         axi_bid_i      : in     slv_03_00;
         --
         axi_awvalid_o  : buffer std_logic;
         axi_awaddr_o   : out    slv_31_00;
         axi_awid_o     : out    slv_03_00;
         axi_awlen_o    : out    slv_07_00;
         axi_awburst_o  : out    slv_01_00;
         axi_wvalid_o   : buffer std_logic;
         axi_wdata_o    : out    slv_31_00;    -- 32b
         axi_wstrb_o    : out    slv_31_00;
         axi_wlast_o    : buffer std_logic;
         axi_bready_o   : buffer std_logic
         );
end entity dcache_axi;


architecture rtl of dcache_axi is

  signal req_q          : std_logic;                -- request start request

  signal req_cnt_q      : slv_t(mem_len_i'range);   -- write burst tracking
  signal req_cnt_0_w    : std_logic;

  signal wr_ack_w       : std_logic;
  signal wr_accept_w    : std_logic;
  signal early_wr_ack_q : std_logic;

  signal rd_accept_w    : std_logic;

  signal mem_addr_q     : slv_31_02;
  signal mem_wdata_q    : slv_31_00;


  type st_t is (IDLE, RD, WR);
  signal st_q : st_t;

begin



  p_st : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        req_cnt_q      <= (others => '0');                        -- 8b
        st_q           <= IDLE;
        mem_addr_q     <= (others => '0');
        mem_wdata_q    <= (others => '0');
        req_q          <= '0';
        early_wr_ack_q <= '0';
      else
        early_wr_ack_q <= '0';
        case st_q is
          when IDLE   =>
            mem_addr_q   <= mem_addr_i(31 downto 2);
            mem_wdata_q  <= mem_wdata_i;
            --
            if (mem_wr_i = '1' and mem_accept_o = '1') then       -- start write
              req_q       <= '1';
              st_q      <= WR;
              req_cnt_q <= mem_len_i;
            end if;
            --
            if (mem_rd_i = '1' and mem_accept_o = '1') then       -- start read
              req_q       <= '1';
              --
              st_q      <= RD;
              req_cnt_q <= mem_len_i;
            end if;
          when WR     =>
            early_wr_ack_q <= axi_wready_i and req_cnt_0_w;
            if (wr_accept_w  = '1') then
              mem_addr_q   <= mem_addr_i(31 downto 2);
              mem_wdata_q  <= mem_wdata_i;
            end if;
            if (axi_awready_i = '1') then
              req_q <= '0';
            end if;
            if (axi_wready_i = '1' and not_zero(req_cnt_q)) then
              req_cnt_q <= dec1u(req_cnt_q);         --  - X"01"
            end if;
            if (axi_bvalid_i  = '1' and axi_bready_o = '1') then  -- wr done
              st_q <= IDLE;
            end if;
          when RD     =>
            if (rd_accept_w = '1') then
              mem_addr_q    <= mem_addr_i(31 downto 2);
            end if;
            if (axi_arready_i = '1') then
              req_q <= '0';
            end if;
            if (axi_rvalid_i  = '1' and axi_rlast_i  = '1') then  -- wr done
              st_q <= IDLE;
            end if;
          when others =>
            st_q <= IDLE;
        end case;
      end if;
    end if;
  end process p_st;



  p_wr : process (st_q, mem_addr_q, mem_wdata_q,
                  axi_wready_i, axi_bvalid_i, early_wr_ack_q,
                  req_q, req_cnt_q, req_cnt_0_w)
  begin
    axi_awvalid_o <= '0';
    axi_wvalid_o  <= '0';
    axi_awaddr_o  <= (others => '0');
    axi_wdata_o   <= (others => '0');
    axi_awid_o    <= (others => '0');
    axi_awlen_o   <= (others => '0');
    axi_wlast_o   <= '0';                     -- max length of burst = 32-bytes
    wr_ack_w      <= '0';
    wr_accept_w   <= '1';
    case st_q is
      when IDLE     =>
        null;
      when WR       =>
        if (req_q = '1') then                 -- start of wr
          axi_awid_o    <= AXI_ID;
          axi_awvalid_o <= '1';
          axi_awlen_o   <= req_cnt_q;
          axi_awaddr_o  <= mem_addr_q & "00"; -- ! ! non-incrementing address
        end if;
        axi_wvalid_o    <= '1';
        axi_wdata_o     <= mem_wdata_q;
        wr_ack_w        <= early_wr_ack_q;    -- early ack for write
        axi_wlast_o     <= axi_wready_i and req_cnt_0_w;
        --
        if (axi_wready_i = '0') then          -- stall if down strem fifo full
          wr_accept_w   <= '0';
        end if;
        if (axi_bvalid_i = '1') then          -- stall between evict-read/vict
          wr_accept_w   <= '0';
        end if;
        --
      when RD  =>
        wr_accept_w     <= '0';
      when others   =>
        null;
    end case;
  end process p_wr;


  req_cnt_0_w   <= is_zero(req_cnt_q);

  axi_wstrb_o    <= X"0000000F";      -- 32b
  axi_awburst_o  <= "01";             -- INCR type only
  axi_bready_o   <= '1';



  p_rd : process (st_q, mem_addr_q, axi_rvalid_i, req_q, req_cnt_q)
  begin
    axi_arvalid_o <= '0';
    axi_araddr_o  <= (others => '0');
    axi_arid_o    <= (others => '0');
    axi_arlen_o   <= (others => '0');
    rd_accept_w   <= '1';
    case st_q is
      when IDLE    =>
        null;
      when RD      =>
        if (req_q = '1') then           -- start of rd
          axi_arid_o    <= AXI_ID;
          axi_arvalid_o <= '1';
          axi_araddr_o  <= mem_addr_q & "00";
          axi_arlen_o   <= req_cnt_q;
        end if;
        rd_accept_w   <= axi_rvalid_i;
      when WR  =>
        rd_accept_w <= '0';
      when others  =>
        null;
    end case;
  end process p_rd;



  axi_arburst_o  <= "01";      -- INCR type only
  axi_rready_o   <= '1';

  mem_rdata_o    <= axi_rdata_i;
  mem_accept_o   <= rd_accept_w or wr_accept_w;
  mem_ack_o      <= axi_rvalid_i or wr_ack_w;
  mem_error_o    <= not_zero(axi_rresp_i) when (axi_rvalid_i = '1') else '0';


end rtl;
