-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- ambit translate off
-- Mem model for simulation and FPGA synthesis (USE lib model in asic synthesis)

-- - - - - - - -
--  libraries
-- - - - - - - -
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.rv_pkg.all;

entity ram_itag is
  generic ( DATA_W             : natural range 8 to 32 := 16;
            ADDR_W             : natural range 4 to 32 := 8;
            ODELAY             : time := 3.5 ns;    -- clk to dout delay
            INITIALIZE_TO_ZERO : std_logic := '0' );
  port ( clk     : in  std_logic;
         xreset  : in  std_logic;                   -- NC now in FPGA synthesis
         sel0    : in  std_logic;                   -- sense delay
         sel1    : in  std_logic;                   -- sense delay
         --  inputs
         rd      : in  std_logic;
         wr      : in  std_logic;
         addr    : in  slv_t(ADDR_W-1 downto 0);
         din     : in  slv_t(DATA_W-1 downto 0);
         --  outputs
         dout    : out slv_t(DATA_W-1 downto 0);
         --  test
         xpd     : in  std_logic;                   -- act low matrix pdown
         se      : in  std_logic;
         si      : in  std_logic;
         scout   : out std_logic );
end entity ram_itag;


architecture behav of ram_itag is

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- single port RAM
  -- mode: read OR write
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  type cachet_t is array (0 to 2**ADDR_W-1) of slv_t(DATA_W-1 downto 0);
  signal ram           : cachet_t;
  signal ram_read_q    : slv_t(DATA_W-1 downto 0);

  -- ambit translate on
  -- pragma translate_off
  signal initialized_w : std_logic := '0';
  signal scan_chain    : slv_t(2+ADDR_W+DATA_W-1 downto 0);
  -- mem output delay from clk rise edge
  signal odelay_w : time;
  -- pragma translate_on
  -- ambit translate off

begin

  -- ambit translate on
  -- pragma translate_off
  p_scano : process (clk)
  begin
    if (clk'event and clk = '1') then
      if (se = '1') then
        scan_chain <= si & scan_chain(scan_chain'left downto 1);
      else
        scan_chain <= rd & wr & addr & din;
      end if;
    end if;
  end process p_scano;
  --
  scout <= scan_chain(0);
  --
  p_delay : process (sel0, sel1)
  begin
    odelay_w <= ODELAY - to_uint(sel1) * 0.4 ns - to_uint(sel0) * 0.2 ns;
  end process p_delay;
  -- pragma translate_on
  -- ambit translate off


  -- synchronous write
  p_ram : process (clk)
  begin
    -- ambit translate on
    -- pragma translate_off
    if (initialized_w = '0' and INITIALIZE_TO_ZERO = '1') then
      ram <= (others =>(others => '0'));
      initialized_w <= '1';
    end if;
    -- pragma translate_on
    -- ambit translate off
    if (clk'event and clk = '1') then
      if ( wr = '1' ) then
        ram(to_uint(addr)) <= din;
      end if;
      if ( rd = '1' ) then
        ram_read_q <=
          -- ambit translate on
          -- pragma translate_off
          (others => 'U') after 0.2 ns,             -- some hold time
          -- pragma translate_on
          -- ambit translate off
          ram(to_uint(addr))
          -- ambit translate on
          -- pragma translate_off
          after odelay_w
          -- pragma translate_on
          -- ambit translate off
          ;
        --
        -- ambit translate on
        -- pragma translate_off
        if (xpd /= '1') then
          ram_read_q <= (others => 'U');
        end if;
        -- pragma translate_on
        -- ambit translate off
        --
      end if;
    end if;
    -- ambit translate on
    -- pragma translate_off
    assert (not (rd = '1' and wr = '1' ))
      report "ERROR: Read and Write are not allowed at the same cycle"
      severity FAILURE;
    -- pragma translate_on
    -- ambit translate off
  end process p_ram;

  dout <= ram_read_q;

end behav;
-- ambit translate on




-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- ambit translate off
-- Mem model for simulation and FPGA synthesis (USE lib model in asic synthesis)

-- - - - - - - -
--  libraries
-- - - - - - - -
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.rv_pkg.all;

entity ram_idata is
  generic ( DATA_W             : natural range 8 to 32 := 16;
            ADDR_W             : natural range 4 to 32 := 8;
            ODELAY             : time := 4.5 ns;    -- clk to dout delay
            INITIALIZE_TO_ZERO : std_logic := '0' );
  port ( clk     : in  std_logic;
         xreset  : in  std_logic;                   -- NC now in FPGA synthesis
         sel0    : in  std_logic;                   -- sense delay
         sel1    : in  std_logic;                   -- sense delay
         --  inputs
         rd      : in  std_logic;
         wr      : in  std_logic;
         addr    : in  slv_t(ADDR_W-1 downto 0);
         din     : in  slv_t(DATA_W-1 downto 0);
         --  outputs
         dout    : out slv_t(DATA_W-1 downto 0);
         --  test
         xpd     : in  std_logic;                   -- act low matrix pdown
         se      : in  std_logic;
         si      : in  std_logic;
         scout   : out std_logic );
end entity ram_idata;


architecture behav of ram_idata is

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- single port RAM
  -- mode: read OR write
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  type cachet_t is array (0 to 2**ADDR_W-1) of slv_t(DATA_W-1 downto 0);
  signal ram           : cachet_t;
  signal ram_read_q    : slv_t(DATA_W-1 downto 0);

  -- ambit translate on
  -- pragma translate_off
  signal initialized_w : std_logic := '0';
  signal scan_chain    : slv_t(1+1+ADDR_W+DATA_W-1 downto 0);
  -- mem output delay from clk rise edge
  signal odelay_w : time;
  -- pragma translate_on
  -- ambit translate off

begin

  -- ambit translate on
  -- pragma translate_off
  p_scano : process (clk)
  begin
    if (clk'event and clk = '1') then
      if (se = '1') then
        scan_chain <= si & scan_chain(scan_chain'left downto 1);
      else
        scan_chain <= rd & wr & addr & din;
      end if;
    end if;
  end process p_scano;
  --
  scout <= scan_chain(0);
  --
  p_delay : process (sel0, sel1)
  begin
    odelay_w <= ODELAY - to_uint(sel1) * 0.4 ns - to_uint(sel0) * 0.2 ns;
  end process p_delay;
  -- pragma translate_on
  -- ambit translate off


  -- synchronous write
  p_ram : process (clk)
  begin
    -- ambit translate on
    -- pragma translate_off
    if (initialized_w = '0' and INITIALIZE_TO_ZERO = '1') then
      ram <= (others =>(others => '0'));
      initialized_w <= '1';
    end if;
    -- pragma translate_on
    -- ambit translate off
    if (clk'event and clk = '1') then
      if ( wr = '1' ) then
        ram(to_uint(addr)) <= din;
      end if;
      if ( rd = '1' ) then
        ram_read_q <=
          -- ambit translate on
          -- pragma translate_off
          (others => 'U') after 0.2 ns,             -- some hold time
          -- pragma translate_on
          -- ambit translate off
          ram(to_uint(addr))
          -- ambit translate on
          -- pragma translate_off
          after odelay_w
          -- pragma translate_on
          -- ambit translate off
          ;
        --
        -- ambit translate on
        -- pragma translate_off
        if (xpd /= '1') then
          ram_read_q <= (others => 'U');
        end if;
        -- pragma translate_on
        -- ambit translate off
        --
      end if;
    end if;
    -- ambit translate on
    -- pragma translate_off
    assert (not (rd = '1' and wr = '1' ))
      report "ERROR: Read and Write are not allowed at the same cycle"
      severity FAILURE;
    -- pragma translate_on
    -- ambit translate off
  end process p_ram;

  dout <= ram_read_q;

end behav;
-- ambit translate on



-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.rv_pkg.all;

entity icache is
  generic ( AXI_ID      : slv_03_00 := X"8";
            WAYS        : natural range 1 to 8  := 2;   -- Nro of ways 1,2,4,8
            LINE_ADDR_W : natural range 6 to 12 := 8;   -- Line address
            LINE_SIZE_W : natural range 2 to 8  := 5 ); -- Line size (bytes !)
  port ( clk_i             : in  std_logic;
         xrst_i            : in  std_logic;
         --  mem inputs (from core)
         cpu_rd_i          : in  std_logic;
         cpu_flush_i       : in  std_logic;
         cpu_invalidate_i  : in  std_logic;
         cpu_pc_i          : in  slv_31_00;
         --  mem outputs (to core)
         cpu_accept_o      : out std_logic;
         cpu_valid_o       : out std_logic;
         cpu_error_o       : out std_logic;
         cpu_inst_o        : out slv_31_00;
         --  AXI write channel (-UNUSED-)
         --  AXI read channel inputs
         axi_arready_i     : in  std_logic;
         axi_rvalid_i      : in  std_logic;
         axi_rdata_i       : in  slv_31_00;
         axi_rresp_i       : in  slv_01_00;
         axi_rid_i         : in  slv_03_00;
         axi_rlast_i       : in  std_logic;
         --  AXI read channel outputs
         axi_arvalid_o     : out std_logic;
         axi_araddr_o      : out slv_31_00;
         axi_arid_o        : out slv_03_00;             -- AXI_ID,  X"8"
         axi_arlen_o       : out slv_07_00;             -- X"07"
         axi_arburst_o     : out slv_01_00;             -- "01", INCR
         axi_rready_o      : out std_logic              -- '1'
         );
end entity icache;


architecture rtl of icache is

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- Default cache instance is 2 way set associative
  -- The total size is 16KB.
  -- The replacement policy is a limited pseudo random scheme (more than 2-ways)
  -- (between lines, toggling on line thrashing).
  -- The cache is a write back cache, with allocate on read and write.
  --
  --   _________    _________________________________________________________
  --  |31|30|29|...|18|17|16|15|14|13|12|11|10| 9| 8| 7| 6| 5| 4| 3| 2| 1| 0|
  --   ---------    ---------------------------------------------------------
  --  |                              |                       | LINE   |     |
  --  |           TAG ADDR           |    LINE ADDR          | OFFSET | byte|
  --  |------------------------------|-----------------------|--------|-----|
  --        31:13, 2^19 = 512k         12:5, 2^8 = 256         2^3=8   ignore
  --
  --   -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
  --  | WAY-0 default implementation                                    |
  --  
  --  |  addr[12:5] (line)      addr[12:5] & addr[4:2] (line+offset)    |
  --      |                      |
  --  |   |         TAG RAM      |         DATA RAM                     |
  --      |         256 x 21bit  |         256*8 x 32bit
  --  |   | 0x00   ___________   | 0x00   ____________                  |
  --      |------>|_ line 0 __|  |------>|_ data 0 ___|
  --  |   |         \      /     |       |_________+1_|                 |
  --      |                      |       |_________+2_|
  --  |   |                      |       |_________+3_|                 |
  --      |                      |       |_________+4_|
  --  |   |                      |       |_________+5_|                 |
  --      |                      |       |_________+6_|
  --  |   |                      |       |_________+6_|                 |
  --      | 0x01   ___________   |       |_________+7_|
  --  |   |------>|_ line 1 __|  |------>|_ data 7 ___|                 |
  --      |         \      /     |       |_________+1_|
  --  |   |                      |       |_________+2_|                 |
  --      |                      |       |_________+3_|
  --  |   |                      |       |_________+4_|                 |
  --      |                      |       |_________+5_|
  --  |   |                      |       |_________+6_|                 |
  --      | 0x02   ___________   |       |_________+7_|
  --  |   |------>|_ line 2 __|  |------>|_ data 16___|                 |
  --      |         \      /     |       |_________+1_|
  --  |   |                      |       |_________+2_|                 |
  --      |                      |       |_________+3_|
  --  |   |                      |       |_________+4_|                 |
  --      |                      |       |_________+5_|
  --  |   |                      |       |_________+6_|                 |
  --      |                      |       |_________+7_|
  --  |   |       . . . . . . . .| . . . . . . . . . . .                |
  --      | 0xFF   ___________   |       |____________|
  --  |   |------>|_ line 255_|  |------>|_ data 2040_|                 |
  --                \      /             |_________+1_|
  --  |                                  |_________+2_|                 |
  --                                     |_________+3_|
  --  |                                  |_________+4_|                 |
  --                                     |_________+5_|
  --  |                                  |_________+6_|                 |
  --                                     |_________+7_|
  --  |__ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __ __|
  --




  -- Address mapping example:                                  NOTE: OBSOLETE
  --                                                           =====
  --  31          16 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
  -- |--------------|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
  --  +--------------------+  +--------------------+   +------------+
  --  |  Tag address       |  |   Line address     |      Address 
  --  |                    |  |                    |      within line
  --  |                    |  |                    |
  --  |                    |  |                    |- LINE_SIZE_W
  --  |                    |  |- LINE_ADDR_H
  --  |                    |- TAG_ADDR_L
  --  |- 31

  
  -- DATA / TAG defs
  constant TAG_ADDR_W   : natural range 15 to 20:= 32-LINE_ADDR_W-LINE_SIZE_W;
  constant TAG_DATA_W   : natural := TAG_ADDR_W + 1;  -- +1 for valid bit
  -- data address width
  constant DATA_ADDR_W  : natural := LINE_ADDR_W + LINE_SIZE_W - 2;  -- words

  -- number of cache lines (words in tag ram) : 2**LINE_ADDR_W

  constant TAG_ADDR_H   : natural := 31;
  constant TAG_ADDR_L   : natural := LINE_ADDR_W + LINE_SIZE_W;

  constant LINE_ADDR_H  : natural := LINE_ADDR_W + LINE_SIZE_W - 1;
  constant LINE_ADDR_L  : natural := LINE_SIZE_W;



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  states
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- oOne hot decoding
  constant STATE_W         : natural := 3;
  constant STATE_FLUSH     : slv_t(STATE_W-1 downto 0) := "001";  -- 0
  constant STATE_LOOKUP    : slv_t(STATE_W-1 downto 0) := "010";  -- 1
  constant STATE_REFILL    : slv_t(STATE_W-1 downto 0) := "100";  -- 2

  constant STATE_FLUSH_B   : natural := 0;
  constant STATE_LOOKUP_B  : natural := 1;
  constant STATE_REFILL_B  : natural := 2;


  component ram_itag
    -- ambit translate off
    generic ( DATA_W             : natural range 8 to 32 := 16;
              ADDR_W             : natural range 4 to 32 := 8;
              ODELAY             : time := 3.5 ns;  -- clk to dout sim delay
              INITIALIZE_TO_ZERO : std_logic := '0' );
    -- ambit translate on
    port ( clk     : in  std_logic;
           xreset  : in  std_logic;                 -- NC now in FPGA synthesis
           sel0    : in  std_logic;                 -- sense delay
           sel1    : in  std_logic;                 -- sense delay
           --  inputs
           rd      : in  std_logic;
           wr      : in  std_logic;
           addr    : in  slv_t(LINE_ADDR_W-1 downto 0);
           din     : in  slv_t(23 downto 0);        -- TAG_DATA_W-1:0
           --  outputs
           dout    : out slv_t(23 downto 0);        -- TAG_DATA_W-1:0
           --  test
           xpd     : in  std_logic;                 -- act low matrix pdown
           se      : in  std_logic;
           si      : in  std_logic;
           scout   : out std_logic );
  end component;

  component ram_idata
    -- ambit translate off
    generic ( DATA_W             : natural range 8 to 32 := 16;
              ADDR_W             : natural range 4 to 32 := 8;
              ODELAY             : time := 4.5 ns;  -- clk to dout sim delay
              INITIALIZE_TO_ZERO : std_logic := '0' );
    -- ambit translate on
    port ( clk     : in  std_logic;
           xreset  : in  std_logic;                 -- NC now in FPGA synthesis
           sel0    : in  std_logic;                 -- sense delay
           sel1    : in  std_logic;                 -- sense delay
           --  inputs
           rd      : in  std_logic;
           wr      : in  std_logic;
           addr    : in  slv_t(DATA_ADDR_W-1 downto 0);
           din     : in  slv_t(32-1 downto 0);
           --  outputs
           dout    : out slv_t(32-1 downto 0);
           --  test
           xpd     : in  std_logic;                 -- act low matrix pdown
           se      : in  std_logic;
           si      : in  std_logic;
           scout   : out std_logic );
  end component;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  registers / wires
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  type tag_data_t  is array (0 to WAYS-1) of slv_23_00;
  type tag_addr_t  is array (0 to WAYS-1) of slv_t(TAG_ADDR_W-1 downto 0);
  type data_data_t is array (0 to WAYS-1) of slv_31_00;

  -- states
  signal next_state_w       : slv_t(STATE_W-1 downto 0);
  signal state_q            : slv_t(STATE_W-1 downto 0);
  signal old_state_lookup_q : std_logic;

  signal invalidate_q       : std_logic;

  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- lookup validation
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  signal lu_valid_q        : std_logic;
  signal lu_addr_q         : slv_31_00;                      -- 1:0 NC
  signal lu_line_addr_w    : slv_t(LINE_ADDR_W-1 downto 0);  -- lookup line addr
  signal lu_data_addr_w    : slv_t(DATA_ADDR_W-1 downto 0);  -- lookup data addr

  signal flush_addr_q      : slv_t(LINE_ADDR_W-1 downto 0);

  signal tag_rd_w          : std_logic;
  signal tag_wr_w          : slv_t(WAYS-1 downto 0);
  signal tag_mdout_w       : tag_data_t;  -- fixed 24 bit ram, extra bits zero
  signal tag_valid_w       : slv_t(WAYS-1 downto 0);
  signal tag_addr_bits_w   : tag_addr_t;

  signal tag_hit_w         : slv_t(WAYS-1 downto 0);
  signal tag_hit_any_w     : std_logic;

  signal data_write_addr_q : slv_t(DATA_ADDR_W-1 downto 0);
  signal data_rd_w         : std_logic;
  signal data_wr_w         : slv_t(WAYS-1 downto 0);

  signal data_mdout_w      : data_data_t;                    -- from mem

  --  data output mux
  signal instr_w           : slv_31_00;

  signal axi_arvalid_q     : std_logic;
  signal axi_error_q       : std_logic;

  signal cpu_accept_o_w    : std_logic;
  signal cpu_valid_o_w     : std_logic;
  signal axi_araddr_o_w    : slv_t(axi_araddr_o'range);
  signal axi_arvalid_o_w   : std_logic;
  signal axi_rready_o_w    : std_logic;

  signal random_num_q      : slv_03_00;
  signal next_way_w        : slv_t(WAYS-1 downto 0);
  signal next_way_q        : slv_t(WAYS-1 downto 0);

  signal refill_bypass_w   : std_logic;

  -- NOTE: TAG ram is fixed to 24 bits, extra msb bits are zero
  signal tag_maddr_w       : slv_t(LINE_ADDR_W-1 downto 0);
  signal data_maddr_w      : slv_t(DATA_ADDR_W-1 downto 0);
  signal tag_mdin_w        : slv_t(23 downto 0);

  signal tag_mrd_w         : slv_t(WAYS-1 downto 0);
  signal tag_mwr_w         : slv_t(WAYS-1 downto 0);
  signal data_mrd_w        : slv_t(WAYS-1 downto 0);
  signal data_mwr_w        : slv_t(WAYS-1 downto 0);

  -- pragma translate_off
  type state_dbg_t is (FLUSH, LOOKUP, REFILL, xxx);
  signal state_dbg         : state_dbg_t;
  -- pragma translate_on

begin
  
  -- pragma translate_off
  state_dbg <= FLUSH    when (state_q = STATE_FLUSH)  else
               LOOKUP   when (state_q = STATE_LOOKUP) else
               REFILL   when (state_q = STATE_REFILL) else
               xxx;
  -- pragma translate_on

  cpu_accept_o    <= cpu_accept_o_w;
  cpu_valid_o     <= cpu_valid_o_w;
  axi_araddr_o    <= axi_araddr_o_w;
  axi_arvalid_o   <= axi_arvalid_o_w;
  axi_rready_o    <= axi_rready_o_w;
  

  lu_line_addr_w  <= cpu_pc_i(LINE_ADDR_H downto LINE_ADDR_L);
  lu_data_addr_w  <= cpu_pc_i(DATA_ADDR_W+2-1 downto 2);


  p_lookup_valid : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        lu_valid_q   <= '0';
      else
        if (cpu_rd_i = '1' and cpu_accept_o_w = '1') then
          lu_valid_q <= '1';
        elsif (cpu_valid_o_w = '1') then
          lu_valid_q <= '0';
        end if;
      end if;
    end if;
  end process p_lookup_valid;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  lookup address
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  p_lookup_addr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        lu_addr_q   <= (others => '0');    -- 32 bit
      else
        if (cpu_rd_i = '1' and cpu_accept_o_w = '1') then
          lu_addr_q <= cpu_pc_i;
        end if;
      end if;
    end if;
  end process p_lookup_addr;



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  TAG RAMS
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  --  tag RAM address
  --  tag RAM write data
  --  NOTE: tag ram data width is fixed to 24-bits
  p_c_tag_addr : process (flush_addr_q, state_q, lu_addr_q, lu_line_addr_w)
  begin
    --
    tag_maddr_w    <= flush_addr_q;
    tag_mdin_w     <= (others => '0');
    --
    if (state_q(STATE_FLUSH_B) = '1') then              -- cache flush
      tag_maddr_w  <= flush_addr_q;
      tag_mdin_w   <= (others => '0');
    elsif (state_q(STATE_REFILL_B) = '1') then          -- line refill
      tag_maddr_w  <= lu_addr_q(LINE_ADDR_H downto LINE_ADDR_L);
      --
      tag_mdin_w(TAG_ADDR_W)            <= '1';         -- valid bit
      tag_mdin_w(TAG_ADDR_W-1 downto 0) <=
        lu_addr_q(TAG_ADDR_H downto TAG_ADDR_L);
    else                                                -- lookup
      tag_maddr_w  <= lu_line_addr_w;
    end if;
  end process p_c_tag_addr;



  --  tag read signal, common for all WAYs
  tag_rd_w  <= state_q(STATE_LOOKUP_B) or
               ( state_q(STATE_REFILL_B) and axi_rlast_i );

  --  data read signal, common for all WAYs
  data_rd_w <= state_q(STATE_LOOKUP_B);



  --  data RAM address
  p_c_data_addr : process (lu_data_addr_w, state_q, data_write_addr_q)
  begin
    data_maddr_w <= lu_data_addr_w;                     -- lookup
    if (state_q(STATE_REFILL_B) = '1') then             -- line refill
      data_maddr_w <= data_write_addr_q;
    end if;
  end process p_c_data_addr;


 

  g_rams : for ii in 0 to WAYS-1 generate
    --
    --  tag RAMs  (semi synchronous ram for tag addr)
    --
    u_tag : ram_itag
      -- ambit translate off
      generic map ( DATA_W             => 24,
                    ADDR_W             => LINE_ADDR_W,
                    ODELAY             => 3.5 ns,
                    INITIALIZE_TO_ZERO => '0' )
      -- ambit translate on
      port map ( clk      => clk_i,
                 xreset   => xrst_i,
                 sel0     => '0',
                 sel1     => '0',
                 --  inputs
                 rd       => tag_mrd_w(ii),          -- common for all WAYs
                 wr       => tag_mwr_w(ii),
                 addr     => tag_maddr_w,
                 din      => tag_mdin_w,             -- 23:0  cache 17 & 16:0
                 --  outputs
                 dout     => tag_mdout_w(ii),        -- 23:0, cache 17 & 16:0
                 --  test
                 xpd      => '1',
                 se       => '0',
                 si       => '0',
                 scout    => OPEN );


    --
    --  data RAMs  (semi synchronous ram for data)
    --
    u_data : ram_idata
      -- ambit translate off
      generic map ( DATA_W             => 32,
                    ADDR_W             => LINE_ADDR_W + LINE_SIZE_W -2,
                    ODELAY             => 4.5 ns,
                    INITIALIZE_TO_ZERO => '0' )
      -- ambit translate on
      port map ( clk      => clk_i,
                 xreset   => xrst_i,
                 sel0     => '0',
                 sel1     => '0',
                 --  inputs
                 rd       => data_mrd_w(ii),         -- common read enable
                 wr       => data_mwr_w(ii),
                 addr     => data_maddr_w,
                 din      => axi_rdata_i,
                 --  outputs
                 dout     => data_mdout_w(ii),
                 --  test
                 xpd      => '1',
                 se       => '0',
                 si       => '0',
                 scout    => OPEN );



    --  Tag RAM write enable  (cache flush or line refill)
    tag_wr_w(ii)         <= state_q(STATE_FLUSH_B) or
                            ( state_q(STATE_REFILL_B) and next_way_q(ii) and
                              old_state_lookup_q and state_q(STATE_REFILL_B));
    
    --  data RAM write enable (way 0 to WAYS-1)
    data_wr_w(ii)        <=  state_q(STATE_REFILL_B) and
                             axi_rvalid_i and next_way_q(ii);

    tag_valid_w(ii)      <= tag_mdout_w(ii)(TAG_ADDR_W);
    tag_addr_bits_w(ii)  <= tag_mdout_w(ii)(TAG_ADDR_W-1 downto 0);
    tag_hit_w(ii)        <= tag_valid_w(ii) and
                            eq(tag_addr_bits_w(ii),
                               lu_addr_q(TAG_ADDR_H downto TAG_ADDR_L));

    --  data in and rd enable is common for all WAYs
    tag_mrd_w(ii)        <= tag_rd_w;
    tag_mwr_w(ii)        <= tag_wr_w(ii);
    data_mrd_w(ii)       <= data_rd_w;
    data_mwr_w(ii)       <= data_wr_w(ii);
    --
  end generate g_rams;


  tag_hit_any_w <= not_zero(tag_hit_w);



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  DATA RAMS
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  --  data RAM refill write address
  p_data_wr_addr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        data_write_addr_q    <= (others => '0');
      else
        if (state_q(STATE_LOOKUP_B) = '1' and
            next_state_w(STATE_REFILL_B) = '1') then
          data_write_addr_q  <= axi_araddr_o_w(DATA_ADDR_W+2-1 downto 2);
        elsif (state_q(STATE_REFILL_B) = '1' and axi_rvalid_i = '1') then
          data_write_addr_q  <= slv_t(unsigned(data_write_addr_q)+1);
        end if;
      end if;
    end if;
  end process p_data_wr_addr;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  flush counter
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  p_flush_addr : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        flush_addr_q    <= (others => '0');
      else
        flush_addr_q    <= (others => '0');
        --
        if (state_q(STATE_FLUSH_B) = '1') then
          flush_addr_q  <=  slv_t(unsigned(flush_addr_q) + 1);
        elsif (cpu_invalidate_i = '1' and cpu_accept_o_w = '1') then
          --  invalidate specified line
          flush_addr_q  <= lu_line_addr_w;
        end if;
      end if;
    end if;
  end process p_flush_addr;



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  replacement policy
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- using random replacement policy
  p_replace_way : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        next_way_q     <= (others => '0');
        next_way_q(0)  <= '1';             -- ...001
      else
        if (state_q(STATE_LOOKUP_B) = '1' and
            next_state_w(STATE_REFILL_B) = '1') then
          next_way_q   <= next_way_w;
        end if;
      end if;
    end if;
  end process p_replace_way;


  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  --  function to select the way to use for fills when WAYS = 1
  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  g_pick_1w : if ( WAYS = 1 ) generate
    next_way_w <= "1";
  end generate g_pick_1w;


  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  --  function to select the way to use for fills when WAYS = 2
  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  g_pick_2w : if ( WAYS = 2 ) generate
    -- pick_way_2ways
    b_pick_2w : block
      --
      function pick_way ( valid_bits : slv_t(WAYS-1 downto 0);
                          random_num : slv_03_00;
                          curr_way   : slv_t(1 downto 0) )
        return slv_t is  -- [WAYS-1:0]
        variable pick_way_v :  slv_t(WAYS-1 downto 0);
      begin
        if ( valid_bits(0) = '0' ) then
          --  way 0 not occupied so use it
          pick_way_v   := "01";           -- ONE HOT encoding in pick_way
        elsif ( valid_bits(1) = '0' ) then
          --  way 1 not occupied so use it
          pick_way_v   := "10";           -- ONE HOT encoding in pick_way
        else
          pick_way_v  := not curr_way;
        end if;
        return pick_way_v;
      end function pick_way;
      --
    begin
      next_way_w <= pick_way( tag_valid_w, random_num_q, next_way_q );
    end block b_pick_2w;
  end generate g_pick_2w;


  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  --  function to select the way to use for fills when WAYS = 4
  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  g_pick_4w : if ( WAYS = 4 ) generate
    -- pick_way_4ways
    b_pick_4w : block
      --
      function pick_way ( valid_bits : slv_t(WAYS-1 downto 0);
                          random_num : slv_03_00 )
        return slv_t is  -- [WAYS-1:0)
        variable pick_way_v :  slv_t(WAYS-1 downto 0);
      begin
        if ( valid_bits(0) = '0' ) then
          --  way 0 not occupied so use it
          pick_way_v     := "0001";
        elsif ( valid_bits(1) = '0' ) then
          --  way 1 not occupied so use it
          pick_way_v     := "0010";
        elsif ( valid_bits(2) = '0' )  then
          --  way 2 not occupied so use it
          pick_way_v     := "0100";
        elsif ( valid_bits(3) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "1000";
        else
          --  all ways occupied so pick one randomly
          case random_num(3 downto 1) is
            when "000" | "001" =>
              pick_way_v := "0100";
            when "010" | "011" =>
              pick_way_v := "1000";
            when "100" | "101" =>
              pick_way_v := "0001";
            when others =>
              pick_way_v := "0010";
          end case;
        end if;
        return pick_way_v;
      end function pick_way;
      --
    begin
      next_way_w <= pick_way( tag_valid_w, random_num_q );
    end block b_pick_4w;
  end generate g_pick_4w;


  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  --  function to select the way to use for fills when WAYS = 8
  -- = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  g_pick_8w : if (  WAYS /= 1 and  WAYS /= 2 and  WAYS /= 3 and
                    WAYS /= 4 and  WAYS /= 5 and  WAYS /= 6 and
                    WAYS /= 7) generate
    --  pick_way_8ways
    b_pick_8w : block
      --
      function pick_way ( valid_bits : slv_t(WAYS-1 downto 0);
                          random_num : slv_03_00 )
        return slv_t is  -- [WAYS-1:0)
        variable pick_way_v :  slv_t(WAYS-1 downto 0);
      begin
        if ( valid_bits(0) = '0' ) then
          --  way 0 not occupied so use it
          pick_way_v     := "00000001";
        elsif ( valid_bits(1) = '0' ) then
          --  way 1 not occupied so use it
          pick_way_v     := "00000010";
        elsif ( valid_bits(2) = '0' ) then
          --  way 2 not occupied so use it
          pick_way_v     := "00000100";
        elsif ( valid_bits(3) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "00001000";
        elsif ( valid_bits(4) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "00010000";
        elsif ( valid_bits(5) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "00100000";
        elsif ( valid_bits(6) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "01000000";
        elsif ( valid_bits(7) = '0' ) then
          --  way 3 not occupied so use it
          pick_way_v     := "10000000";
        else
          --  all ways occupied so pick one randomly
          case random_num(3 downto 1) is
            when "000"  =>  pick_way_v := "00010000";
            when "001"  =>  pick_way_v := "00100000";
            when "010"  =>  pick_way_v := "01000000";
            when "011"  =>  pick_way_v := "10000000";
            when "100"  =>  pick_way_v := "00000001";
            when "101"  =>  pick_way_v := "00000010";
            when "110"  =>  pick_way_v := "00000100";
            when others =>  pick_way_v := "00001000";
          end case;
        end if;
        return pick_way_v;
      end function pick_way;
      --
    begin
      next_way_w <= pick_way( tag_valid_w, random_num_q );
    end block b_pick_8w;
  end generate g_pick_8w;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  instruction output
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  g_dout_1w : if ( WAYS = 1 ) generate
    instr_w <= data_mdout_w(0);
  end generate g_dout_1w;
  --
  --
  g_dout_2w : if ( WAYS = 2 ) generate
    p_c_inst_mux : process (data_mdout_w, tag_hit_w)
      variable tag_sel_v : slv_01_00;
    begin
      --  case has to be of static type
      tag_sel_v := slv_t(unsigned(tag_hit_w));
      case tag_sel_v is
        when "10"   => instr_w <= data_mdout_w(1);
        when others => instr_w <= data_mdout_w(0);
      end case;
    end process p_c_inst_mux;
  end generate g_dout_2w;
  --
  --
  g_dout_4w : if ( WAYS = 4 ) generate
    p_c_inst_mux : process (data_mdout_w, tag_hit_w)
      variable tag_sel_v : slv_03_00;
    begin
      --  case has to be of static type
      tag_sel_v := slv_t(unsigned(tag_hit_w));
      case tag_sel_v is
        when "1000" => instr_w <= data_mdout_w(3);
        when "0100" => instr_w <= data_mdout_w(2);
        when "0010" => instr_w <= data_mdout_w(1);
        when others => instr_w <= data_mdout_w(0);
      end case;
    end process p_c_inst_mux;
  end generate g_dout_4w;
  --
  --
  g_dout_8w : if ( WAYS = 8 ) generate
    p_c_inst_mux : process (data_mdout_w, tag_hit_w)
      variable tag_sel_v : slv_07_00;
    begin
      --  case has to be of static type
      tag_sel_v := slv_t(unsigned(tag_hit_w));
      case tag_sel_v is
        when "10000000" => instr_w <= data_mdout_w(7);
        when "01000000" => instr_w <= data_mdout_w(6);
        when "00100000" => instr_w <= data_mdout_w(5);
        when "00010000" => instr_w <= data_mdout_w(4);
        when "00001000" => instr_w <= data_mdout_w(3);
        when "00000100" => instr_w <= data_mdout_w(2);
        when "00000010" => instr_w <= data_mdout_w(1);
        when others     => instr_w <= data_mdout_w(0);
      end case;
    end process p_c_inst_mux;
  end generate g_dout_8w;


  --  bypass data and ack core when changing state from refill -> lookup
  --  lookup data valid OR Bypass after last refill
  cpu_valid_o_w <= (lu_valid_q and state_q(STATE_LOOKUP_B) and
                    tag_hit_any_w) or  (refill_bypass_w);


  --  bypass data and ack core during refill
  refill_bypass_w <= state_q(STATE_REFILL_B) and
                     eq(lu_addr_q(LINE_SIZE_W-1 downto 2),
                        data_write_addr_q(LINE_SIZE_W-1-2 downto 0)) and
                     axi_rvalid_i;


  cpu_inst_o <= axi_rdata_i when (refill_bypass_w = '1') else instr_w;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  next state logic
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --
  --                          xrst_i == '0'
  --                              |
  --                              |
  --                             \/
  --                         STATE_FLUSH <--------------
  --                              |                     |
  --  invalidate_q == '1'  OR     |                     |
  --  flush_addr_q == X"FF"       |                     |
  --                             \/                     |
  --                         STATE_LOOKUP --------------
  --                              |    /\         cpu_invalidate_i == '1' OR
  --  lu_valid_q == '1' AND       |     \         cpu_flush_i      == '1'
  --  tag_hit_any_w == '0'        |      \        ( invalidate a line /
  --  ( tried a lookup but        |       \         flush cache )
  --    no match found )         \/        \
  --                         STATE_REFILL   \
  --                              |          |
  --  axi_rvalid_i == '1' AND     `- - - - ->'
  --  axi_rlast_i  == '1
  --
  --

  p_next_state : process (state_q, invalidate_q, flush_addr_q, lu_valid_q,
                          tag_hit_any_w, cpu_invalidate_i, cpu_flush_i,
                          axi_rvalid_i, axi_rlast_i)
  begin
    --
    next_state_w <= state_q;
    --
    case state_q is
      -- - - - - - - - - - - - - - - - - - - - -
      -- STATE_FLUSH
      -- - - - - - - - - - - - - - - - - - - - -
      when STATE_FLUSH =>
        if (invalidate_q = '1') then
          next_state_w  <= STATE_LOOKUP;
        elsif (unsigned(not flush_addr_q) = 0) then   -- all ONES => last addr
          next_state_w  <= STATE_LOOKUP;
        end if;
      -- - - - - - - - - - - - - - - - - - - - -
      -- STATE_LOOKUP
      -- - - - - - - - - - - - - - - - - - - - -
      when STATE_LOOKUP =>
        --  tried a lookup but no match found
        if (lu_valid_q = '1' and  tag_hit_any_w = '0') then
          next_state_w  <= STATE_REFILL;
        --  invalidate a line / flush cache
        elsif (cpu_invalidate_i = '1' or cpu_flush_i = '1') then
          next_state_w  <= STATE_FLUSH;
        end if;
      -- - - - - - - - - - - - - - - - - - - - -
      -- STATE_REFILL
      -- - - - - - - - - - - - - - - - - - - - -
      when STATE_REFILL =>
        --  end of refill
        if (axi_rvalid_i = '1' and  axi_rlast_i = '1') then
          next_state_w  <= STATE_LOOKUP;
        end if;
      when others =>
        -- null;
        next_state_w    <= STATE_FLUSH;
    end case;
  end process p_next_state;



  --  update state
  p_state : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        state_q            <= STATE_FLUSH;
        old_state_lookup_q <= '0';
        random_num_q       <= X"F";
      else
        state_q            <= next_state_w;
        old_state_lookup_q <= state_q(STATE_LOOKUP_B);
        random_num_q       <= random_num_q(2) & random_num_q(1) &
                              random_num_q(0) &
                              ( random_num_q(3) xor random_num_q(2) );
      end if;
    end if;
  end process p_state;


  cpu_accept_o_w <= state_q(STATE_LOOKUP_B) and
                    ( not next_state_w(STATE_REFILL_B) );


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  invalidate
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  p_invalidate : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        invalidate_q     <= '0';
      else
        invalidate_q     <= '0';
        if (cpu_invalidate_i = '1' and cpu_accept_o_w = '1') then
          invalidate_q   <= '1';
        end if;
      end if;
    end if;
  end process p_invalidate;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  AXI request hold
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  p_axi_arvalid : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        axi_arvalid_q     <= '0';
      else
        axi_arvalid_q     <= '0';
        if (axi_arvalid_o_w = '1' and axi_arready_i = '0') then
          axi_arvalid_q   <= '1';
        end if;
      end if;
    end if;
  end process p_axi_arvalid;


  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  AXI error handling
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  p_axi_error : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        axi_error_q     <= '0';
      else
        if (axi_rvalid_i = '1' and axi_rready_o_w = '1' and
            not_zero(axi_rresp_i)) then
          axi_error_q   <= '1';
        elsif (cpu_valid_o_w = '1') then
          axi_error_q   <= '0';
        end if;
      end if;
    end if;
  end process p_axi_error;


  cpu_error_o <= axi_error_q;



  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  -- AXI
  -- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  --  AXI write channel (unused)

  --  AXI read channel
  axi_arvalid_o_w <= axi_arvalid_q or
                     (state_q(STATE_LOOKUP_B) and next_state_w(STATE_REFILL_B));


  -- lu_addr_q(31:LINE_SIZE_W)    &    - 31:5
  -- (LINE_SIZE_W-1 downto 0 => '0')   - "00000"
  axi_araddr_o_w  <= slv_t(shift_left(
    resize(unsigned(lu_addr_q(31 downto LINE_SIZE_W)),
           axi_araddr_o_w'length), LINE_SIZE_W));

  axi_arburst_o   <= "01";    -- INCR
  axi_arid_o      <= AXI_ID;  -- X"8"
  axi_arlen_o     <= slv_t(
    to_unsigned( (2**(LINE_SIZE_W-2) -1 ), axi_arlen_o'length));
  axi_rready_o_w  <= '1';

end rtl;
