-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--
-- Copyright 2025 VLSI Solution Oy
-- SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
--
-- Licensed under the Solderpad Hardware License v 2.1 (the "License");
-- you may not use this file except in compliance with the License, or,
-- at your option, the Apache License version 2.0.
-- You may obtain a copy of the License at
--
--    https://solderpad.org/licenses/SHL-2.1/
--
-- Unless required by applicable law or agreed to in writing, any work
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
--                               RISCV core
--                                 V1.0
--
--
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

-- - - - - - - -
--  libraries
-- - - - - - - -
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.rv_pkg.all;


entity rv_core is
  generic ( BOOT_ADDR      : slv_31_00 := X"80000000";
            CPU_ID         : slv_31_00 := X"00000000";
            ICACHE_AXI_ID  : slv_03_00 := X"8";
            DCACHE_AXI_ID  : slv_03_00 := X"4";
            CACHE_ADDR_L   : slv_31_00 := X"80000000";
            CACHE_ADDR_H   : slv_31_00 := X"8FFFFFFF"
            );
  port ( clk_i            : in  std_logic;
         xrst_i           : in  std_logic;                  -- syncr reset
         xrst_cpu_i       : in  std_logic;                  -- riscv core only
         --
         axi_i_awready_i  : in  std_logic;                  -- 11 AXI_I_I
         axi_i_wready_i   : in  std_logic;
         axi_i_bvalid_i   : in  std_logic;
         axi_i_bresp_i    : in  slv_01_00;
         axi_i_bid_i      : in  slv_03_00;
         axi_i_arready_i  : in  std_logic;
         axi_i_rvalid_i   : in  std_logic;
         axi_i_rdata_i    : in  slv_31_00;
         axi_i_rresp_i    : in  slv_01_00;
         axi_i_rid_i      : in  slv_03_00;
         axi_i_rlast_i    : in  std_logic;
         axi_i_awvalid_o  : out std_logic;                  -- 16 AXI_O_I
         axi_i_awaddr_o   : out slv_31_00;
         axi_i_awid_o     : out slv_03_00;
         axi_i_awlen_o    : out slv_07_00;
         axi_i_awburst_o  : out slv_01_00;
         axi_i_wvalid_o   : out std_logic;
         axi_i_wdata_o    : out slv_31_00;
         axi_i_wstrb_o    : out slv_03_00;
         axi_i_wlast_o    : out std_logic;
         axi_i_bready_o   : out std_logic;
         axi_i_arvalid_o  : out std_logic;
         axi_i_araddr_o   : out slv_31_00;
         axi_i_arid_o     : out slv_03_00;
         axi_i_arlen_o    : out slv_07_00;
         axi_i_arburst_o  : out slv_01_00;
         axi_i_rready_o   : out std_logic;
         --
         axi_d_awready_i  : in  std_logic;                  -- 11 AXI_I_D
         axi_d_wready_i   : in  std_logic;
         axi_d_bvalid_i   : in  std_logic;
         axi_d_bresp_i    : in  slv_01_00;
         axi_d_bid_i      : in  slv_03_00;
         axi_d_arready_i  : in  std_logic;
         axi_d_rvalid_i   : in  std_logic;
         axi_d_rdata_i    : in  slv_31_00;
         axi_d_rresp_i    : in  slv_01_00;
         axi_d_rid_i      : in  slv_03_00;
         axi_d_rlast_i    : in  std_logic;
         axi_d_awvalid_o  : out std_logic;                  -- 16 AXI_O_D
         axi_d_awaddr_o   : out slv_31_00;
         axi_d_awid_o     : out slv_03_00;
         axi_d_awlen_o    : out slv_07_00;
         axi_d_awburst_o  : out slv_01_00;
         axi_d_wvalid_o   : out std_logic;
         axi_d_wdata_o    : out slv_31_00;
         axi_d_wstrb_o    : out slv_03_00;
         axi_d_wlast_o    : out std_logic;
         axi_d_bready_o   : out std_logic;
         axi_d_arvalid_o  : out std_logic;
         axi_d_araddr_o   : out slv_31_00;
         axi_d_arid_o     : out slv_03_00;
         axi_d_arlen_o    : out slv_07_00;
         axi_d_arburst_o  : out slv_01_00;
         axi_d_rready_o   : out std_logic;
         --  vsbus for peripherals
         pp_irq_i         : in  std_logic;                  -- perip interrupt
         pp_addr_o        : out slv_31_00;
         pp_rd_o          : out std_logic;
         pp_rd_data_i     : in  slv_31_00;
         pp_wr_o          : out slv_03_00;
         pp_wr_data_o     : out slv_31_00
         );
end entity rv_core;


architecture rtl of rv_core is

  component rv_cpu
    generic (
      BOOT_ADDR       : slv_31_00 := X"80000000";
      CPU_ID          : slv_31_00 := X"00000000";
      CACHE_ADDR_L    : slv_31_00 := X"80000000";
      CACHE_ADDR_H    : slv_31_00 := X"8FFFFFFF";
      --
      RISCV_M         : std_logic := '1';   -- RV mul / div extension
      RISCV_E         : std_logic := '0';   -- RV rf embedded extension
      RISCV_S_U       : std_logic := '1';   -- supervisor/user privilege levels
  --  RISCV_I         : std_logic := '1';   -- always enabled when no RISCV_E !
  --  RISCV_Zicsr     : std_logic := '1';   -- always enabled !
  --  RISCV_Zifencei  : std_logic := '1';   -- always enabled !
      --
      SUPPORT_MMU         : std_logic := '1';
      SUPPORT_LOAD_BYPASS : std_logic := '1';
      SUPPORT_MUL_BYPASS  : std_logic := '1';
      EXTRA_DECODE_STAGE  : std_logic := '1' );
    port ( clk_i              : in  std_logic;
           xrst_i             : in  std_logic;    -- syncr reset
           --   interrupt request from irq ctrl
           irq_i              : in  std_logic;    -- external inter
           tmrirq_i           : in  std_logic;    -- riscv timer inter
           --  icache inputs
           imem_accept_i      : in  std_logic;
           imem_valid_i       : in  std_logic;
           imem_error_i       : in  std_logic;
           imem_data_i        : in  slv_31_00;
           --  icache outputs
           imem_rd_o          : out std_logic;
           imem_pc_o          : out slv_31_00;
           imem_flush_o       : out std_logic;
           imem_invalidate_o  : out std_logic;
           --  dcache/peripheral inputs
           dmem_data_rd_i     : in  slv_31_00;
           dmem_accept_i      : in  std_logic;
           dmem_ack_i         : in  std_logic;
           dmem_error_i       : in  std_logic;
           dmem_resp_tag_i    : in  slv_10_00;
           --  dcache/peripheral outputs
           dmem_rd_o          : out std_logic;
           dmem_wr_o          : out slv_03_00;
           dmem_addr_o        : out slv_31_00;
           dmem_data_wr_o     : out slv_31_00;
           dmem_cacheable_o   : out std_logic;
           dmem_req_tag_o     : out slv_10_00;
           dmem_invalidate_o  : out std_logic;
           dmem_wb_o          : out std_logic;
           dmem_flush_o       : out std_logic );
  end component;


  component icache
    generic ( AXI_ID      : slv_03_00 := X"8";
              WAYS        : natural range 1 to 8  := 2;   -- Nro of ways 1,2,4,8
              LINE_ADDR_W : natural range 6 to 12 := 8;   -- Line address
              LINE_SIZE_W : natural range 2 to 8  := 5 ); -- Line size (bytes !)
    port ( clk_i             : in  std_logic;
           xrst_i            : in  std_logic;             -- syncr reset
           --  cache inputs (from core)
           cpu_rd_i          : in  std_logic;
           cpu_flush_i       : in  std_logic;
           cpu_invalidate_i  : in  std_logic;
           cpu_pc_i          : in  slv_31_00;
           --  cache outputs (to core)
           cpu_accept_o      : out std_logic;
           cpu_valid_o       : out std_logic;
           cpu_error_o       : out std_logic;
           cpu_inst_o        : out slv_31_00;
           --  NO AXI write channel
           --  AXI read channel, inputs
           axi_arready_i     : in  std_logic;
           axi_rvalid_i      : in  std_logic;
           axi_rdata_i       : in  slv_31_00;
           axi_rresp_i       : in  slv_01_00;
           axi_rid_i         : in  slv_03_00;
           axi_rlast_i       : in  std_logic;
           --  AXI read channel, outputs
           axi_arvalid_o     : out std_logic;
           axi_araddr_o      : out slv_31_00;
           axi_arid_o        : out slv_03_00;
           axi_arlen_o       : out slv_07_00;
           axi_arburst_o     : out slv_01_00;
           axi_rready_o      : out std_logic );
  end component;


  component dcache
    generic (AXI_ID      : slv_03_00 := X"4";
             WAYS        : natural range 1 to 8  := 2;   -- Nro of ways 1,2,4,8
             LINE_ADDR_W : natural range 6 to 12 := 8;   -- Line address
             LINE_SIZE_W : natural range 2 to 8  := 5 ); -- Line size, byte addr
    port ( clk_i             : in  std_logic;
           xrst_i            : in  std_logic;            -- syncr reset
           --
           cpu_addr_i        : in  slv_31_00;
           cpu_data_wr_i     : in  slv_31_00;
           cpu_rd_i          : in  std_logic;
           cpu_wr_i          : in  slv_03_00;
           cpu_req_tag_i     : in  slv_10_00;
           cpu_invalidate_i  : in  std_logic;
           cpu_writeback_i   : in  std_logic;
           cpu_flush_i       : in  std_logic;
           cpu_data_rd_o     : out slv_31_00;
           cpu_accept_o      : out std_logic;
           cpu_ack_o         : out std_logic;
           cpu_error_o       : out std_logic;
           cpu_resp_tag_o    : out slv_10_00;
           --
           axi_awready_i     : in  std_logic;
           axi_wready_i      : in  std_logic;
           axi_bvalid_i      : in  std_logic;
           axi_bresp_i       : in  slv_01_00;
           axi_bid_i         : in  slv_03_00;
           axi_arready_i     : in  std_logic;
           axi_rvalid_i      : in  std_logic;
           axi_rdata_i       : in  slv_31_00;
           axi_rresp_i       : in  slv_01_00;
           axi_rid_i         : in  slv_03_00;
           axi_rlast_i       : in  std_logic;
           --
           axi_awvalid_o     : out std_logic;
           axi_awaddr_o      : out slv_31_00;
           axi_awid_o        : out slv_03_00;
           axi_awlen_o       : out slv_07_00;
           axi_awburst_o     : out slv_01_00;
           axi_wvalid_o      : out std_logic;
           axi_wdata_o       : out slv_31_00;
           axi_wstrb_o       : out slv_03_00;
           axi_wlast_o       : out std_logic;
           axi_bready_o      : out std_logic;
           axi_arvalid_o     : out std_logic;
           axi_araddr_o      : out slv_31_00;
           axi_arid_o        : out slv_03_00;
           axi_arlen_o       : out slv_07_00;
           axi_arburst_o     : out slv_01_00;
           axi_rready_o      : out std_logic );
  end component;


  signal icache_rd_w          : std_logic;
  signal icache_pc_w          : slv_31_00;
  signal icache_inst_x01_w    : slv_31_00;
  signal icache_inst_w        : slv_31_00;
  signal icache_accept_w      : std_logic;
  signal icache_valid_w       : std_logic;
  signal icache_flush_w       : std_logic;
  signal icache_error_w       : std_logic;
  signal icache_invalidate_w  : std_logic;

  signal core_d_cacheable_w   : std_logic;
  signal core_d_accept_w      : std_logic;
  signal core_d_ack_w         : std_logic;
  signal core_d_rd_w          : std_logic;                     -- core_d rd ena
  signal core_d_wr_w          : slv_03_00;                     -- core_d wr ena
  signal core_d_addr_w        : slv_31_00;
  signal core_d_data_rd_w     : slv_31_00;
  signal core_d_data_wr_w     : slv_31_00;
  signal core_d_resp_tag_w    : slv_10_00;
  signal core_d_req_tag_w     : slv_10_00;
  signal core_d_wb_w          : std_logic;
  signal core_d_error_w       : std_logic;
  signal core_d_flush_w       : std_logic;
  signal core_d_invalidate_w  : std_logic;

  signal dcache_accept_w      : std_logic;
  signal dcache_ack_w         : std_logic;
  signal dcache_cached_rd_w   : std_logic;                     -- dcache rd ena
  signal dcache_cached_wr_w   : slv_03_00;                     -- dcache wr ena
  signal dcache_data_rd_w     : slv_31_00;
  signal dcache_resp_tag_w    : slv_10_00;
  signal dcache_error_w       : std_logic;

  --  perip control
  signal core_rd_q            : std_logic;
  signal core_wr_q            : slv_03_00;
  signal core_addr_q          : slv_31_00;
  signal core_wr_data_q       : slv_31_00;
  signal core_d_cacheable_q   : std_logic;

  signal pp_access_w          : std_logic;
  signal pp_ack_w             : std_logic;

begin


  p_pp_retime : process (clk_i)
  begin
    if (clk_i'event and clk_i = '1') then
      if (xrst_i = '0') then
        core_rd_q          <= '0';
        core_wr_q          <= X"0";                     -- 3:0   --! 3:1 NC
        core_addr_q        <= X"00000000";              -- 31:0
        core_wr_data_q     <= X"00000000";              -- 31:0
        core_d_cacheable_q <= '0';
      else
        if (core_d_accept_w = '1') then
          core_rd_q          <= core_d_rd_w;
          core_wr_q          <= core_d_wr_w;            -- 3:0   --! 3:1 NC
          core_addr_q        <= core_d_addr_w;          -- 31:0
          core_wr_data_q     <= core_d_data_wr_w;       -- 31:0
          core_d_cacheable_q <= core_d_cacheable_w;
        end if;
      end if;
    end if;
  end process p_pp_retime;


  --
  -- NOTE: peripheral bus has a one clock cycle delay in all operations
  --
  pp_access_w  <= not core_d_cacheable_q;
  pp_addr_o    <= core_addr_q;
  pp_rd_o      <= core_rd_q and pp_access_w;
  pp_wr_o      <= core_wr_q and (pp_access_w & pp_access_w &
                                 pp_access_w & pp_access_w);
  pp_wr_data_o <= core_wr_data_q;
  -- always ack after one clock
  pp_ack_w     <= (not core_d_cacheable_q) and
                  (core_rd_q or ne( core_wr_q,"0000"));



  -- dcache read/write enebles
  dcache_cached_rd_w <= core_d_rd_w and core_d_cacheable_w;
  dcache_cached_wr_w <= core_d_wr_w and ( core_d_cacheable_w &
                                          core_d_cacheable_w &
                                          core_d_cacheable_w &
                                          core_d_cacheable_w );



  --
  -- MUXED dcache/perip inputs to core
  --
  core_d_data_rd_w <= dcache_data_rd_w when (core_d_cacheable_q = '1') else
                      pp_rd_data_i;

  core_d_ack_w     <= pp_ack_w when (core_d_cacheable_q = '0') else
                      dcache_ack_w;

  core_d_resp_tag_w <= dcache_resp_tag_w;


  core_d_accept_w <= dcache_accept_w  and  '1';    -- perips always ready
  core_d_error_w  <= dcache_error_w   or   '0';    -- perips never error



  u_cpu : rv_cpu
    generic map ( BOOT_ADDR       => BOOT_ADDR,
                  CPU_ID          => CPU_ID,
                  CACHE_ADDR_L    => CACHE_ADDR_L,
                  CACHE_ADDR_H    => CACHE_ADDR_H,
                  --
                  RISCV_M         => '1',   -- RV mul / div extension
                  RISCV_E         => '0',   -- RV rf embedded extension
                  RISCV_S_U       => '1',   -- supervisor/user privilege levels
               -- RISCV_I         => '1',   -- always enabled when no RISCV_E !
               -- RISCV_Zicsr     => '1',   -- always enabled !
               -- RISCV_Zifencei  => '1',   -- always enabled !
                  --
                  SUPPORT_MMU         => '1',
                  SUPPORT_LOAD_BYPASS => '1',
                  SUPPORT_MUL_BYPASS  => '1',
                  EXTRA_DECODE_STAGE  => '1' )
    port map ( clk_i              => clk_i,
               xrst_i             => xrst_cpu_i,
               --  interrupt request from irq ctrl
               irq_i              => pp_irq_i,   -- external inter from irq_ctrl
               tmrirq_i           => '0',        -- riscv timer inter
               -- icache inputs
               imem_accept_i      => icache_accept_w,
               imem_valid_i       => icache_valid_w,
               imem_error_i       => icache_error_w,
               imem_data_i        => icache_inst_w,
               --  icache outputs
               imem_rd_o          => icache_rd_w,
               imem_pc_o          => icache_pc_w,
               imem_flush_o       => icache_flush_w,
               imem_invalidate_o  => icache_invalidate_w,
               --  dcache/peripheral inputs
               dmem_data_rd_i     => core_d_data_rd_w,  -- MUXED dcache/perip
               dmem_accept_i      => core_d_accept_w,
               dmem_ack_i         => core_d_ack_w,      -- MUXED dcache/perip
               dmem_error_i       => core_d_error_w,
               dmem_resp_tag_i    => core_d_resp_tag_w,
               --  dcache/peripheral outputs
               dmem_rd_o          => core_d_rd_w,       -- dcache/perip rd ena
               dmem_wr_o          => core_d_wr_w,       -- dcache/perip wr ena
               dmem_addr_o        => core_d_addr_w,
               dmem_data_wr_o     => core_d_data_wr_w,
               dmem_cacheable_o   => core_d_cacheable_w,
               dmem_req_tag_o     => core_d_req_tag_w,
               dmem_invalidate_o  => core_d_invalidate_w,
               dmem_wb_o          => core_d_wb_w,
               dmem_flush_o       => core_d_flush_w );


  icache_inst_w <=
    -- pragma translate_off
    to_stdlogicvector(to_bitvector(
      -- pragma translate_on
      icache_inst_x01_w
      -- pragma translate_off
      ))
    -- pragma translate_on
    ;



  u_icache : icache
    -- DEFAULT SETUP:
    generic map ( AXI_ID       => ICACHE_AXI_ID, -- X"8",
                  WAYS         => 2,             -- Number of ways 1,2,4,8
                  LINE_ADDR_W  => 8,             -- Line address
                  LINE_SIZE_W  => 5 )            -- Line size (bytes !)
    port map ( clk_i             => clk_i,
               xrst_i            => xrst_cpu_i,
               --  cache inputs (from core)
               cpu_rd_i          => icache_rd_w,
               cpu_flush_i       => icache_flush_w,
               cpu_invalidate_i  => icache_invalidate_w,
               cpu_pc_i          => icache_pc_w,
               --  cache outputs (to core)
               cpu_accept_o      => icache_accept_w,
               cpu_valid_o       => icache_valid_w,
               cpu_error_o       => icache_error_w,
               cpu_inst_o        => icache_inst_x01_w,
               --  NO AXI write channel
               --  AXI read channel, inputs
               axi_arready_i     => axi_i_arready_i,
               axi_rvalid_i      => axi_i_rvalid_i,
               axi_rdata_i       => axi_i_rdata_i,
               axi_rresp_i       => axi_i_rresp_i,
               axi_rid_i         => axi_i_rid_i,
               axi_rlast_i       => axi_i_rlast_i,
               --  AXI read channel, outputs
               axi_arvalid_o     => axi_i_arvalid_o,
               axi_araddr_o      => axi_i_araddr_o,
               axi_arid_o        => axi_i_arid_o,
               axi_arlen_o       => axi_i_arlen_o,
               axi_arburst_o     => axi_i_arburst_o,
               axi_rready_o      => axi_i_rready_o );

--  write channel is NC
--  axi_i_awready_i
--  axi_i_wready_i
--  axi_i_bvalid_i
--  axi_i_bresp_i
--  axi_i_bid_i
    axi_i_awvalid_o <= '0';
    axi_i_awaddr_o  <= (others => '0');
    axi_i_awid_o    <= (others => '0');
    axi_i_awlen_o   <= (others => '0');
    axi_i_awburst_o <= (others => '0');
    axi_i_wvalid_o  <= '0';
    axi_i_wdata_o   <= (others => '0');
    axi_i_wstrb_o   <= (others => '0');
    axi_i_wlast_o   <= '0';
    axi_i_bready_o  <= '0';




  u_dcache : dcache
    -- DEFAULT SETUP:
    generic map ( AXI_ID       => DCACHE_AXI_ID, -- X"4"
                  WAYS         => 2,             -- Number of ways 1,2,4,8
                  LINE_ADDR_W  => 8,             -- Line address
                  LINE_SIZE_W  => 5 )            -- Line size (bytes !)
    port map ( clk_i             => clk_i,
               xrst_i            => xrst_cpu_i,
               --  core inputs
               cpu_addr_i        => core_d_addr_w,
               cpu_data_wr_i      => core_d_data_wr_w,
               cpu_rd_i          => dcache_cached_rd_w,
               cpu_wr_i          => dcache_cached_wr_w,
               cpu_req_tag_i     => core_d_req_tag_w,
               cpu_invalidate_i  => core_d_invalidate_w,
               cpu_writeback_i   => core_d_wb_w,
               cpu_flush_i       => core_d_flush_w,
               --  core outputs
               cpu_data_rd_o     => dcache_data_rd_w,
               cpu_accept_o      => dcache_accept_w,
               cpu_ack_o         => dcache_ack_w,
               cpu_error_o       => dcache_error_w,
               cpu_resp_tag_o    => dcache_resp_tag_w,
               --  AXI4 inputs
               axi_awready_i     => axi_d_awready_i,
               axi_wready_i      => axi_d_wready_i,
               axi_bvalid_i      => axi_d_bvalid_i,
               axi_bresp_i       => axi_d_bresp_i,
               axi_bid_i         => axi_d_bid_i,
               axi_arready_i     => axi_d_arready_i,
               axi_rvalid_i      => axi_d_rvalid_i,
               axi_rdata_i       => axi_d_rdata_i,
               axi_rresp_i       => axi_d_rresp_i,
               axi_rid_i         => axi_d_rid_i,
               axi_rlast_i       => axi_d_rlast_i,
               --  AXI4 outputs
               axi_awvalid_o     => axi_d_awvalid_o,
               axi_awaddr_o      => axi_d_awaddr_o,
               axi_awid_o        => axi_d_awid_o,
               axi_awlen_o       => axi_d_awlen_o,
               axi_awburst_o     => axi_d_awburst_o,
               axi_wvalid_o      => axi_d_wvalid_o,
               axi_wdata_o       => axi_d_wdata_o,
               axi_wstrb_o       => axi_d_wstrb_o,
               axi_wlast_o       => axi_d_wlast_o,
               axi_bready_o      => axi_d_bready_o,
               axi_arvalid_o     => axi_d_arvalid_o,
               axi_araddr_o      => axi_d_araddr_o,
               axi_arid_o        => axi_d_arid_o,
               axi_arlen_o       => axi_d_arlen_o,
               axi_arburst_o     => axi_d_arburst_o,
               axi_rready_o      => axi_d_rready_o );

end rtl;
